"""
Output and Ansi art unicode helpers for x/84, https://github.com/jquast/x84
"""
import re
import math
import warnings
import textwrap

from x84.bbs.session import getterminal, getsession
from x84.bbs.wcwidth import wcswidth

ANSI_PIPE = re.compile(r'\|(\d{2,3})')
ANSI_COLOR = re.compile(r'\033\[(\d{2,3})m')
ANSI_RIGHT = re.compile(r'\033\[(\d{1,4})C')
ANSI_CODEPAGE = re.compile(r'\033[\(\)][AB012]')
ANSI_WILLMOVE = re.compile(r'\033\[[HJuABCDEF]')
ANSI_WONTMOVE = re.compile(r'\033\[[sm]')


class AnsiWrapper(textwrap.TextWrapper):
    # pylint: disable=C0111
    #         Missing docstring
    def _wrap_chunks(self, chunks):
        """
        ANSI-string safe varient of wrap_chunks,
        exception of movement sequences.
        """
        lines = []
        if self.width <= 0:
            raise ValueError("invalid width %r (must be > 0)" % self.width)
        chunks.reverse()
        while chunks:
            cur_line = []
            cur_len = 0
            if lines:
                indent = self.subsequent_indent
            else:
                indent = self.initial_indent
            width = self.width - len(indent)
            if self.drop_whitespace and chunks[-1].strip() == '' and lines:
                del chunks[-1]
            while chunks:
                chunk_len = len(Ansi(chunks[-1]))
                if cur_len + chunk_len <= width:
                    cur_line.append(chunks.pop())
                    cur_len += chunk_len
                else:
                    break
            if chunks and len(Ansi(chunks[-1])) > width:
                self._handle_long_word(chunks, cur_line, cur_len, width)
            if (self.drop_whitespace
                    and cur_line
                    and cur_line[-1].strip() == ''):
                del cur_line[-1]
            if cur_line:
                lines.append(indent + u''.join(cur_line))
        return lines
AnsiWrapper.__doc__ = textwrap.TextWrapper.__doc__


def ansiwrap(ucs, width=70, **kwargs):
    """Wrap a single paragraph of Unicode Ansi sequences,
    returning a list of wrapped lines.
    """
    return AnsiWrapper(width=width, **kwargs).wrap(ucs)


def echo(ucs):
    """
    Display unicode terminal sequence.
    """
    session = getsession()
    if not isinstance(ucs, unicode):
        warnings.warn('non-unicode: %r' % (ucs,), UnicodeWarning, 2)
        return session.write(ucs.decode('iso8859-1'))
    return session.write(ucs)


def timeago(secs, precision=0):
    """
    timago(float[,int])

    Pass a duration of time and return human readable shorthand, fe.

    asctime(126.32) -> ' 2m 6s',
    asctime(70.9999, 2) -> ' 1m 10.99s'
    """
    # split by days, mins, hours, secs
    years = weeks = days = mins = hours = 0
    mins, secs = divmod(secs, 60)
    hours, mins = divmod(mins, 60)
    days, hours = divmod(hours, 24)
    weeks, days = divmod(days, 7)
    years, weeks = divmod(weeks, 52)
    ((num1, num2), (label1, label2)) = (
        ((years, weeks), (u'y', u'w')) if years >= 1.0 else
        ((weeks, days), (u'w', u'd')) if weeks >= 1.0 else
        ((days, hours), (u'd', u'h')) if days >= 1.0 else
        ((hours, mins), (u'h', u'm')) if hours >= 1.0 else
        ((mins, secs), (u'm', u's')))
    return (u'%2d%s%2.*f%s' % (num1, label1, precision, num2, label2,))


class Ansi(unicode):
    """
    This unicode variation understands the effect of ansi sequences of
    printable length, as well as double-wide east asian characters on
    terminals, properly implementing .rjust, .ljust, .center, and .len.

    Other ANSI functions also provided as methods.
    """
    # pylint: disable=R0904,R0924
    #         Too many public methods (45/20)
    #         Badly implemented Container, implements __getitem__,
    #           __len__ but not __delitem__, __setitem__.

    # this is really bad; this is old kludge dating as far back as 2002 from a
    # prior period when ansi was implemented without blessings, shoehorned into
    # this unicode-derived class ..

    def __len__(self):
        """
        Return the printed length of a string that contains (some types) of
        ansi sequences. Although accounted for, strings containing sequences
        such as cls() will not give accurate returns. backspace, delete, and
        double-wide east-asian
        """
        # 'nxt' points to first *ch beyond current ansi sequence, if any.
        # 'width' is currently estimated display length.
        nxt, width = 0, 0

        # i regret the heavy re-instantiation of Ansi() ..
        for idx in range(0, unicode.__len__(self)):
            width += Ansi(self[idx:]).anspadd()
            if idx == nxt:
                nxt = idx + Ansi(self[idx:]).seqlen()
            if nxt <= idx:
                ucs = self[idx]
                if getsession().encoding == 'cp437':
                    wide = 1
                else:
                    # 'East Asian Fullwidth' and 'East Asian Wide' characters
                    # can take 2 cells, see
                    # http://www.unicode.org/reports/tr11/
                    # http://www.gossamer-threads.com/lists/python/bugs/972834
                    # we just use wcswidth, since that is what terminal
                    # client implementors seem to be using ..
                    wide = wcswidth(ucs)

                # TODO
                # my own NVT addition: allow -1 to be added to width when
                # 127 and 8 are used (BACKSPACE, DEL); as well as \x0f .!?
#                assert wide != -1 or ucs in (u'\b',
#                                            unichr(127),
#                                            unichr(15)), (
#                    'indeterminate length %r in %r' % (self[idx], self))
                width += wide if wide != -1 else 0
                nxt = idx + Ansi(self[idx:]).seqlen() + 1
        return width

    def ljust(self, width):
        return self + u' ' * (max(0, width - self.__len__()))
    ljust.__doc__ = unicode.ljust.__doc__

    def rjust(self, width):
        return u' ' * (max(0, width - self.__len__())) + self
    rjust.__doc__ = unicode.rjust.__doc__

    def center(self, width):
        split = max(0.0, float(width) - self.__len__()) / 2
        return (u' ' * (max(0, int(math.floor(split)))) + self
                + u' ' * (max(0, int(math.ceil(split)))))
    center.__doc__ = unicode.center.__doc__

    def wrap(self, width, indent=u''):
        """
        A.wrap(width) -> unicode

        Like textwrap.wrap, but honor existing linebreaks and understand
        printable length of a unicode string that contains ANSI sequences.

        Always returns single unicode string,
        rows are seperated by NVT \\r\\n newlines
        """
        lines = []
        for line in self.splitlines():
            if line.strip():
                for wrapped in ansiwrap(line, width, subsequent_indent=indent):
                    lines.append(wrapped)
            else:
                lines.append(u'')
        return '\r\n'.join(lines)

    def is_movement(self):
        """
        S.is_movement() -> bool

        Returns True if string S begins with a known terminal sequence that is
        unhealthy for padding, that is, its effects on the terminal window
        position are indeterminate.
        """
        # pylint: disable=R0911,R0912
        #        Too many return statements (20/6)
        #        Too many branches (23/12)
        slen = unicode.__len__(self)
        if 0 == slen:
            return False
        elif self[0] != unichr(27):
            return False
        elif self[1] == u'c':
            # reset
            return True
        elif slen < 3:
            # unknown
            return False
        elif ANSI_CODEPAGE.match(self):
            return False
        elif (self[0], self[1], self[2]) == (u'#', u'8'):
            # 'fill the screen'
            return True
        elif ANSI_WILLMOVE.match(self):
            return True
        elif ANSI_WONTMOVE.match(self):
            return False
        elif slen < 4:
            # unknown
            return False
        elif self[2] == '?':
            # CSI + '?25(h|l)' # show|hide
            ptr2 = 3
            while (self[ptr2].isdigit()):
                ptr2 += 1
            if not self[ptr2] in u'hl':
                # ? followed illegaly, UNKNOWN
                return False
            return False
        elif self[2] in ('(', ')'):
            # CSI + '\([AB012]' # set G0/G1
            assert self[3] in (u'A', 'B', '0', '1', '2',)
            return False
        elif not self[2].isdigit():
            # illegal nondigit in seq
            return False
        ptr2 = 2
        while (self[ptr2].isdigit()):
            ptr2 += 1
        # multi-attribute SGR '[01;02(..)'(m|H)
        n_tries = 0
        while ptr2 < slen and self[ptr2] == ';' and n_tries < 64:
            n_tries += 1
            ptr2 += 1
            try:
                while (self[ptr2].isdigit()):
                    ptr2 += 1
                if self[ptr2] == 'H':
                    # 'H' pos,
                    return True
                elif self[ptr2] == 'm':
                    # 'm' color;attr
                    return False
                elif self[ptr2] == ';':
                    # multi-attribute SGR
                    continue
            except IndexError:
                # out-of-range in multi-attribute SGR
                return False
            # illegal multi-attribtue SGR
            return False
        if ptr2 >= slen:
            # unfinished sequence, hrm ..
            return False
        elif self[ptr2] in u'ABCDEFGJKSTH':
            # single attribute,
            # up, down, right, left, bnl, bpl,
            # pos, cls, cl, pgup, pgdown
            return True
        elif self[ptr2] == 'm':
            # normal
            return False
        # illegal single value, UNKNOWN
        return False

    def seqfill(self, encode_pipe=False):
        """
         S.seqfill() -> unicode

           Pad string S, previously filled with the terminal
           "cursor right" sequence, ``<ESC>[<N>C``, with space character ' ',
           also used as 'erase'. Ansi art can also be compressed in this way,
           by replacing with padded u' 's, we can scroll such artwork
           bi-directionally or within pager windows without 'bleeding' at the
           cost of extra bytes.

           When encode_pipe is True, color sequences are replaced with
           user-editable pipe sequences following the guidelines of MCI
           codes, http://wiki.mysticbbs.com/mci_codes. When False (default),
           colors are stripped entirely.
        """
        ptr = 0
        rstr = u''
        for idx in range(0, unicode.__len__(self)):
            seq_left = Ansi(self[idx:]).seqlen()
            if seq_left:
                ptr = idx + seq_left
                rstr += u' ' * (Ansi(self[idx:]).anspadd())
            elif seq_left and Ansi(self[idx:]).is_movement():
                ptr = idx + seq_left
                if encode_pipe:
                    match = ANSI_COLOR.match(self[idx:])
                    if match:
                        # http://wiki.mysticbbs.com/mci_codes
                        value = int(self[match.start():match.end() + 1]) - 30
                        if value >= 0 and value <= 60:
                            rstr += u'|%02d' % (value,)
            elif ptr <= idx:
                rstr += self[idx]
        return rstr

    def anspadd(self):
        """
         S.anspadd() -> integer

        Returns int('nn') in CSI sequence \\033[nnC for use with replacing
        ansi.right(nn) with printable characters. prevents bleeding in
        scrollable windows. Otherwise 0
        """
        right = ANSI_RIGHT.match(self)
        if right is not None:
            return int(right.group(1))
        return 0

    def encode_pipe(self):
        """
        S.encode_pipe() <==> S.seqfill(encode_pipe=True)
        """
        return self.seqfill(encode_pipe=True)

    def decode_pipe(self):
        """
        S.decode_pipe() -> unicode

        Return new terminal sequence, replacing 'pipe codes', such as u'|03'
        with this terminals equivalent attribute sequence.
        """
        term = getterminal()
        ucs = u''
        ptr = 0
        match = None
        for match in ANSI_PIPE.finditer(self):
            ucs_value = match.group(1)
            # allow escaping using a second pipe
            if match.start() and self[match.start() - 1] == '|':
                continue
            # 07 -> 7
            while ucs_value.startswith('0'):
                ucs_value = ucs_value[1:]
            int_value = 0 if 0 == len(ucs_value) else int(ucs_value, 10)
            assert int_value >= 0 and int_value <= 256
            # colors 0-7 and 16-256 are as-is term.color()
            # special accomidations for 8-15, some termcaps are ok
            # with term.color(11), whereas others have trouble, help
            # out by using dim color and bold attribute instead.
            attr = u''
            if int_value <= 7 or int_value >= 16:
                attr = term.normal + term.color(int_value)
            elif int_value <= 15:
                attr = term.normal + term.color(int_value - 8) + term.bold
            ucs += self[ptr:match.start()] + attr
            ptr = match.end()
        if match is None:
            ucs = self
        else:
            ucs += self[match.end():]
        ptr = 0
        return ''.join((ucs, term.normal))

    def seqlen(self):
        """
        S.is_sequence() -> integer

        Returns non-zero for string S that begins with an ansi sequence, with
        value of bytes until sequence is complete. Use as a 'next' pointer to
        skip past sequences.
        """
        # pylint: disable=R0911,R0912
        #        Too many return statements (19/6)
        #        Too many branches (22/12)
        slen = unicode.__len__(self)
        if 0 == slen:
            return 0  # empty string
        elif self[0] != unichr(27):
            return 0  # not a sequence
        elif 1 == slen:
            return 0  # just esc,
        elif self[1] == u'c':
            return 2  # reset
        elif 2 == slen:
            return 0  # not a sequence
        elif (self[1], self[2]) == (u'#', u'8'):
            return 3  # fill screen (DEC)
        elif ANSI_CODEPAGE.match(self) or ANSI_WONTMOVE.match(self):
            return 3
        elif ANSI_WILLMOVE.match(self):
            return 4
        elif self[1] == '[':
            # all sequences are at least 4 (\033,[,0,m)
            if slen < 4:
                # not a sequence !?
                return 0
            elif self[2] == '?':
                # CSI + '?25(h|l)' # show|hide
                ptr2 = 3
                while (self[ptr2].isdigit()):
                    ptr2 += 1
                if not self[ptr2] in u'hl':
                    # ? followed illegaly, UNKNOWN
                    return 0
                return ptr2 + 1
            # SGR
            elif self[2].isdigit():
                ptr2 = 2
                while (self[ptr2].isdigit()):
                    ptr2 += 1
                    if ptr2 == unicode.__len__(self):
                        return 0

                # multi-attribute SGR '[01;02(..)'(m|H)
                while self[ptr2] == ';':
                    ptr2 += 1
                    if ptr2 == unicode.__len__(self):
                        return 0
                    try:
                        while (self[ptr2].isdigit()):
                            ptr2 += 1
                    except IndexError:
                        return 0
                    if self[ptr2] in u'Hm':
                        return ptr2 + 1
                    elif self[ptr2] == ';':
                        # multi-attribute SGR
                        continue
                    # 'illegal multi-attribute sgr'
                    return 0
                # single attribute SGT '[01(A|B|etc)'
                if self[ptr2] in u'ABCDEFGJKSTHm':
                    # single attribute,
                    # up/down/right/left/bnl/bpl,pos,cls,cl,
                    # pgup,pgdown,color,attribute.
                    return ptr2 + 1
                # illegal single value
                return 0
            # illegal nondigit
            return 0
        # unknown...
        return 0



"""
Ansi Windowing package for x/84, http://github.com/jquast/x84
"""
from x84.bbs.cp437 import CP437TABLE
from x84.bbs.session import getterminal, getsession
from x84.bbs.output import Ansi

GLYPHSETS = {
    'ascii': {
        'top-left': u'+',
        'bot-left': u'+',
        'top-right': u'+',
        'bot-right': u'+',
        'left-vert': u'|',
        'right-vert': u'|',
        'top-horiz': u'-',
        'bot-horiz': u'-', },
    'thin': {
        'top-left': CP437TABLE[unichr(218)],
        'bot-left': CP437TABLE[unichr(192)],
        'top-right': CP437TABLE[unichr(191)],
        'bot-right': CP437TABLE[unichr(217)],
        'left-vert': CP437TABLE[unichr(179)],
        'right-vert': CP437TABLE[unichr(179)],
        'top-horiz': CP437TABLE[unichr(196)],
        'bot-horiz': CP437TABLE[unichr(196)],
    },
}


class AnsiWindow(object):
    """
    The AnsiWindow base class provides position-relative window drawing
    routines to terminal interfaces, such as pager windows, editors, and
    lightbar lists.
    """
    # pylint: disable=R0902
    #        Too many instance attributes (8/7)

    def __init__(self, height, width, yloc, xloc):
        """
        Construct an ansi window. Its base purpose is to provide
        window-relativie positions using the pos() method.
        """
        self.height = height
        self.width = width
        self.yloc = yloc
        self.xloc = xloc
        self._xpadding = 1
        self._ypadding = 1
        self._alignment = 'left'
        self._moved = False
        self.glyphs = dict()
        self.colors = dict()
        self.init_theme()
        term = getterminal()
        assert self.isinview(), (
            'AnsiWindow(height=%s, width=%s, yloc=%s, xloc=%s)'
            ' not in viewport Terminal(height=%s, width=%s)'
            % (height, width, yloc, xloc, term.height, term.width))

    def init_theme(self):
        """
        This initializer sets glyphs and colors appropriate for a "theme",
        override this method to create a common color and graphic set.
        """
        session = getsession()
        term = getterminal()
        self.colors['normal'] = term.normal
        if term.number_of_colors != 0:
            self.colors['border'] = term.cyan
        # start with default 'ascii'
        self.glyphs = GLYPHSETS['ascii'].copy()
        # PC-DOS 'thin' on smart terminals
        if session.env.get('TERM') != 'unknown':
            self.glyphs = GLYPHSETS['thin'].copy()

    @property
    def xpadding(self):
        """
        Horizontal padding of window border
        """
        return self._xpadding

    @xpadding.setter
    def xpadding(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._xpadding = value

    @property
    def ypadding(self):
        """
        Veritcal padding of window border
        """
        return self._ypadding

    @ypadding.setter
    def ypadding(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._ypadding = value

    @property
    def alignment(self):
        """
        Justification of text content. One of 'left', 'right', or 'center'.
        """
        return self._alignment

    @alignment.setter
    def alignment(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        assert value in ('left', 'right', 'center')
        self._alignment = value

    def align(self, text, width=None):
        """
        justify Ansi text alignment property and width. When None (default),
        the visible width after padding is used.
        """
        return (Ansi(text).rjust
                if self.alignment == 'right'
                else Ansi(text).ljust
                if self.alignment == 'left'
                else Ansi(text).center
                )(width if width is not None else self.visible_width)

    @property
    def visible_height(self):
        """
        The visible height of the editor window after vertical padding.
        """
        return self.height - (self.ypadding * 2)

    @property
    def visible_width(self):
        """
        Visible width of lightbar after accounting for horizontal padding.
        """
        return self.width - (self.xpadding * 2)

    def resize(self, height=None, width=None, yloc=None, xloc=None):
        """
        Adjust window dimensions.
        """
        if height is not None:
            self.height = height
        if width is not None:
            self.width = width
        if yloc is not None:
            self.yloc = yloc
        if xloc is not None:
            self.xloc = xloc

    def isinview(self):
        """
        Returns True if window is in view of the terminal window.
        """
        term = getterminal()
        return (self.xloc >= 0
                and self.xloc + self.width <= term.width
                and self.yloc >= 0
                and self.yloc + self.height <= term.height)

    def iswithin(self, win):
        """
        Returns True if our window is within the bounds of window
        """
        return (self.yloc >= win.yloc
                and self.yloc + self.height <= win.yloc + win.height
                and self.xloc >= win.xloc
                and self.xloc + self.width <= win.xloc + win.width)

    def willfit(self, win):
        """
        Returns True if target window is within our bounds
        """
        return (win.yloc >= self.yloc
                and win.yloc + win.height <= self.yloc + self.height
                and win.xloc >= self.xloc
                and win.xloc + win.w <= self.xloc + self.width)

    def pos(self, yloc=None, xloc=None):
        """
        Returns terminal sequence to move cursor to window-relative position.
        """
        term = getterminal()
        return term.move((yloc if yloc is not None else 0) + self.yloc,
                         (xloc if xloc is not None else 0) + self.xloc)

    def title(self, ansi_text):
        """
        Returns sequence that positions and displays unicode sequence
        'ansi_text' at the title location of the window.
        """
        xloc = self.width / 2 - min(len(Ansi(ansi_text)) / 2, self.width / 2)
        return self.pos(0, max(0, xloc)) + ansi_text

    def footer(self, ansi_text):
        """
        Returns sequence that positions and displays unicode sequence
        'ansi_text' at the bottom edge of the window.
        """
        xloc = self.width / 2 - min(len(Ansi(ansi_text)) / 2, self.width / 2)
        return self.pos(max(0, self.height - 1), max(0, xloc)) + ansi_text

    def border(self):
        """
        Return a unicode sequence suitable for drawing a border of this window
        using self.colors 'border', 'normal' and glyphs:
        'top-left', 'top-horiz', 'top-right', 'left-vert', 'right-vert',
        'bot-left', 'bot-horiz', and 'bot-right'.
        """
        # pylint: disable=R0912
        #        Too many branches (17/12)
        topright = self.glyphs.get('top-right', u'*')
        thoriz = self.glyphs.get('top-horiz', u'-') * (max(0, self.width - 2))
        topleft = self.glyphs.get('top-left', u'/')
        leftvert = self.glyphs.get('left-vert', u'|')
        rightvert = self.glyphs.get('right-vert', u'|')
        botleft = self.glyphs.get('bot-left', u'\\')
        bhoriz = self.glyphs.get('bot-horiz', u'-') * (max(0, self.width - 2))
        botright = self.glyphs.get('bot-right', u'/')
        rstr = u''
        for row in range(0, self.height):
            for col in range(0, self.width):
                if (col == 0) or (col == self.width - 1):
                    rstr += (self.pos(row, col) + topleft
                             if row == 0 and col == 0 else
                             self.pos(row, col) + botleft
                             if row == self.height - 1 and col == 0 else
                             self.pos(row, col) + topright
                             if row == 0 else
                             self.pos(row, col) + botright
                             if row == self.height - 1 else
                             self.pos(row, col) + leftvert
                             if col == 0 else
                             self.pos(row, col) + rightvert
                             if col == self.width - 1 else
                             u'')
                elif (row == 0):
                    # top row (column 1)
                    if thoriz == u'':
                        if topright != u'':
                            # prepare for top-right, (horiz skipped)
                            rstr += self.pos(row, max(0, self.width - 1))
                    else:
                        rstr += thoriz
                    rstr += topright
                    break
                elif (row == self.height - 1):
                    # bottom row (column 1)
                    if bhoriz == u'':
                        if botright != u'':
                            # prepare for bot-right, (horiz skipped)
                            rstr += self.pos(row, max(0, self.width - 1))
                    else:
                        # horizontal line
                        rstr += bhoriz
                    # bot-right
                    rstr += botright
                    break
        return (self.colors.get('border', u'') + rstr +
                self.colors.get('normal', u''))

    def erase_border(self):
        """
        Return a unicode sequence suitable for erasing
        the border of this window.
        """
        save = self.glyphs.copy()
        for glyph in ('top-left', 'top-horiz', 'top-right',
                      'left-vert', 'right-vert',
                      'bot-left', 'bot-horiz', 'bot-right'):
            if 0 != len(self.glyphs.get(glyph, u'')):
                self.glyphs[glyph] = self.glyphs.get('era):se', u' ')
        ucs = self.border()
        self.glyphs = save
        return ucs

    def erase(self):
        """
        Return a unicode sequence suitable for erasing
        this window (includes border).
        """
        return u''.join([self.pos(y, 0)
                         + (self.glyphs.get('erase', u' ') * self.width)
                         for y in range(self.height)
                         ])

    def clear(self):
        """
        Return a unicode sequence suitable for erasing
        the contents of this window (border remains).
        """
        return u''.join([self.pos(self.ypadding + yloc, self.xpadding)
                         + (self.glyphs.get('erase', u' ')
                            * self.visible_width)
                         for yloc in range(self.visible_height)
                         ])

    @property
    def moved(self):
        """
        Returns True if movement has occured, can be reset to
        False by caller.
        """
        return self._moved

    @moved.setter
    def moved(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        assert type(value) is bool
        self._moved = value



"""
Userbase record access for x/84, https://github.com/jquast/x84/
"""
import logging
from x84.bbs.dbproxy import DBProxy

# pylint: disable=C0103
#        Invalid name "logger" for type constant
logger = logging.getLogger()
digestpw = None
GROUPDB = 'groupbase'
USERDB = 'userbase'


def list_users():
    """
    Returns all user handles.
    """
    return [handle.decode('utf8')
            for handle in DBProxy(USERDB).keys()]


def get_user(handle):
    """
    Returns User record, keyed by handle.
    """
    return DBProxy(USERDB)[handle]


def find_user(handle):
    """
    Given handle, discover and return matching database key case insensitively.
    The returned value may not be equal to the argument, or None if not found.
    """
    for key in DBProxy(USERDB).keys():
        if handle.lower() == key.decode('utf8').lower():
            return key


def digestpw_bcrypt(password, salt=None):
    """
    Password digest using bcrypt (optional)
    """
    import bcrypt
    if not salt:
        salt = bcrypt.gensalt()
    return salt, bcrypt.hashpw(password, salt)


def digestpw_internal(password, salt=None):
    """
    Password digest using regular python libs
    """
    import hashlib
    import base64
    import os
    if not salt:
        salt = base64.b64encode(os.urandom(32))
    digest = salt + password
    for _count in range(0, 100000):
        # pylint: disable=E1101
        #         Module 'hashlib' has no 'sha256'
        digest = hashlib.sha256(digest).hexdigest()
    return salt, digest


def digestpw_plaintext(password, salt=None):
    """
    No password digest, just store the passwords in plain text
    """
    if not salt:
        salt = 'none'
    return salt, password


def digestpw_init(password_digest):
    """
    Set which password digest routine to use
    """
    # pylint: disable=W0603
    #         Using the global statement
    global digestpw
    if password_digest == 'bcrypt':
        digestpw = digestpw_bcrypt
    elif password_digest == 'internal':
        digestpw = digestpw_internal
    elif password_digest == 'plaintext':
        digestpw = digestpw_plaintext
    else:
        assert False, ('Invalid value for "system.password_digest"')


class Group(object):
    """
    A simple group record object with properties name', and 'members'.
    Use methods 'add' and 'remove' to add and remove users by handle,
    and 'save' to persist to disk (and other bbs sessions).
    """

    def __init__(self, name, members=()):
        """
        Initialize a group of name and members
        """
        self._name = name
        self._members = set(members)

    @property
    def name(self):
        """
        Name of this group.
        """
        return self._name

    @name.setter
    def name(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._name = value

    @property
    def members(self):
        """
        Members of this group as user handles.
        """
        return self._members

    def add(self, handle):
        """
        Add user to group.
        """
        logger.info("Group('%s').add('%s')", self.name, handle)
        self._members.add(handle)

    def remove(self, handle):
        """
        Remove user from group.
        """
        logger.info("Group('%s').remove('%s')", self.name, handle)
        self._members.remove(handle)

    def save(self):
        """
        Save group record to database.
        """
        DBProxy(GROUPDB)[self.name] = self

    def delete(self):
        """
        Delete group record from database, and from .groups of any users.
        """
        udb = DBProxy(USERDB)
        for chk_user in self.members:
            user = udb[chk_user]
            if self.name in user.groups:
                user.group_del(self.name)
                user.save()
        del DBProxy(GROUPDB)[self.name]


class User(object):
    """
    A simple user record object with setter and getter properties, 'handle',
    'location', 'email', 'password', 'groups', 'calls', 'lastcall', and 'plan'.
    """
    # pylint: disable=R0902,R0924
    #        Too many instance attributes (8/7)
    #        Badly implemented Container, implements
    #          __delitem__, __getitem__, __setitem__
    #          but not __len__

    def __init__(self, handle=u'anonymous'):
        """
        Initialize a user record, using handle
        """
        self._handle = handle
        self._password = (None, None)
        self._location = u''
        self._email = u''
        self._groups = set()
        self._calls = 0
        self._lastcall = 0
        self._plan = u''

    @property
    def handle(self):
        """
        U.handle() --> unicode

        User handle, also the database key.
        """
        return self._handle

    @handle.setter
    def handle(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._handle = value

    @property
    def password(self):
        """
        U.password() --> tuple

        Returns encrypted form of password as tuple (salt, hash).
        If a password has not yet been set, it is (None, None).
        """
        return self._password

    @password.setter
    def password(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        from x84.bbs import ini
        if ini.CFG.getboolean('system', 'pass_ucase'):
            # facebook and mystic storage style, i wouldn't
            # recommend it though.
            self._password = digestpw(value.upper())
        else:
            self._password = digestpw(value)
        logger.info("set password for user '%s'.", self.handle)

    def auth(self, try_pass):
        """
        U.auth(unicode) --> boolean

        To authenticate, assert auth(try_pass) == True.
        """
        assert type(try_pass) is unicode
        assert len(try_pass) > 0
        assert self.password != (None, None), ('account is without password')
        salt = self.password[0]
        return (self.password == digestpw(try_pass, salt) or
                self.password == digestpw(try_pass.upper(), salt))

    def __setitem__(self, key, value):
        # pylint: disable=C0111,
        #        Missing docstring
        if self.handle == 'anonymous':
            logger.debug("set attr %r not possible for 'anonymous'", key)
            return
        adb = DBProxy(USERDB, 'attrs')
        adb.acquire()
        if not self.handle in adb:
            adb[self.handle] = dict([(key, value), ])
        else:
            attrs = adb[self.handle]
            attrs.__setitem__(key, value)
            adb[self.handle] = attrs
        adb.release()
        logger.info("set attr %r for user '%s'.", key, self.handle)
    __setitem__.__doc__ = dict.__setitem__.__doc__

    def get(self, key, default=None):
        # pylint: disable=C0111,
        #        Missing docstring
        adb = DBProxy(USERDB, 'attrs')
        adb.acquire()
        if not self.handle in adb:
            logger.debug(
                '%r GET %r: default; missing attrs.', self.handle, key)
            val = default
        else:
            attrs = adb[self.handle]
            if not key in attrs:
                logger.debug('%r GET %r: default', self.handle, key)
                val = default
            else:
                logger.debug('%r GET %r%s.' % (
                    self.handle, key,
                    ' (size: %d)' % (len(attrs[key]),)
                    if hasattr(attrs[key], '__len__')
                    else '(1)'))
                val = attrs[key]
        adb.release()
        return val
    get.__doc__ = dict.get.__doc__

    def __getitem__(self, key):
        # pylint: disable=C0111,
        #        Missing docstring
        return DBProxy(USERDB, 'attrs')[self.handle][key]
    __getitem__.__doc__ = dict.__getitem__.__doc__

    def __delitem__(self, key):
        # pylint: disable=C0111,
        #        Missing docstring
        uadb = DBProxy(USERDB, 'attrs')
        uadb.acquire()
        attrs = uadb[self.handle]
        if key in attrs:
            attrs.__delitem__(key)
            uadb[self.handle] = attrs
        uadb.release()
        logger.info("del attr %r for user '%s'.", key, self.handle)
    __delitem__.__doc__ = dict.__delitem__.__doc__

    @property
    def groups(self):
        """
        U.groups() --> set

        Set of Group records user is a member of.
        """
        return self._groups

    def group_add(self, group):
        """
        Add user to group.
        """
        return self._groups.add(group)

    def group_del(self, group):
        """
        Remove user from group.
        """
        return self._groups.remove(group)

    def save(self):
        """
        (re-)Save user record to databases. Changes to user record to not
        automaticly persist. a call to the .save method must be done.
        """
        assert type(self._handle) is unicode, ('handle must be unicode')
        assert len(self._handle) > 0, ('handle must be non-zero length')
        assert (None, None) != self._password, ('password must be set')
        assert self._handle != u'anonymous', (
            'anonymous user may not be saved.')
        udb = DBProxy(USERDB)
        udb.acquire()
        if 0 == len(udb) and self.is_sysop is False:
            logger.warn('%s: First new user becomes sysop.', self.handle)
            self.group_add(u'sysop')
        udb[self.handle] = self
        adb = DBProxy(USERDB, 'attrs')
        adb.acquire()
        if not self.handle in adb:
            adb[self.handle] = dict()
        adb.release()
        self._apply_groups(DBProxy(GROUPDB))
        udb.release()
        logger.info("saved user '%s'.", self.handle)

    def delete(self):
        """
        Remove user record from database, and as a member of any groups.
        """
        gdb = DBProxy(GROUPDB)
        for gname in self._groups:
            group = gdb[gname]
            if self.handle in group.members:
                group.remove(self.handle)
                group.save()
        del DBProxy(USERDB)[self.handle]
        logger.info("deleted user '%s'.", self.handle)

    @property
    def is_sysop(self):
        """
        U.is_sysop --> boolean

        Returns True if user is in u'sysop' group.
        """
        return u'sysop' in self._groups

    @property
    def lastcall(self):
        """
        U.lastcall() --> float

        Time last called, time.time() epoch-formatted.
        """
        return self._lastcall

    @lastcall.setter
    def lastcall(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._lastcall = value

    @property
    def calls(self):
        """Legacy, number of times user has 'called' this board."""
        return self._calls

    @calls.setter
    def calls(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._calls = value

    @property
    def location(self):
        """
        Legacy, used as a geographical location, group names, etc.
        """
        return self._location

    @location.setter
    def location(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._location = value

    @property
    def email(self):
        """
        E-mail address. May be used for password resets.
        """
        return self._email

    @email.setter
    def email(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._email = value

    @property
    def plan(self):
        """
        Unix .plan contents, the original blogosphere.
        """
        return self._plan

    @plan.setter
    def plan(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._plan = value

    def _apply_groups(self, gdb):
        """
        Inspect all groupbase members and enforce referential integrity.
        """
        gdb.acquire()
        for chk_grp in self._groups:
            if not chk_grp in gdb:
                gdb[chk_grp] = Group(chk_grp, set([self.handle]))
                logger.info("created group '%s' for user '%s'.",
                        chk_grp, self.handle)
            # ensure membership in existing groups
            group = gdb[chk_grp]
            if not self.handle in group.members:
                group.add(self.handle)
                group.save()
        for gname, group in gdb.items():
            if gname not in self._groups and self.handle in group.members:
                group.remove(self.handle)
                group.save()
        gdb.release()



"""
 Main menu script for x/84, http://github.com/jquast/x84
"""


def refresh():
    """ Refresh main menu. """
    # pylint: disable=R0914
    #         Too many local variables
    from x84.bbs import getsession, getterminal, echo, Ansi, showcp437, ini
    import os
    session, term = getsession(), getterminal()
    session.activity = u'Main menu'
    artfile = os.path.join(os.path.dirname(__file__), 'art', 'main.asc')
    echo(u''.join((
        u'\r\n\r\n',
        term.blue(u'/'.rjust(term.width / 2)), term.bold_black(u'/ '),
        term.bold('x'), term.bold_blue('/'), term.bold('84'), u' ',
        'MAiN MENU',
        u'\r\n')))
    for line in showcp437(artfile):
        echo(line)
    echo(u'\r\n\r\n')
    entries = [
        ('b', 'bS NEXUS'),
        ('l', 'ASt CAllS'),
        ('o', 'NE liNERS'),
        ('w', "hO'S ONliNE"),
        ('n', 'EWS'),
        ('c', 'hAt'),
        ('!', 'ENCOdiNG'),
        ('t', 'EtRiS'),
        ('s', 'YS. iNfO'),
        ('f', 'ORECASt'),
        ('e', 'dit PROfilE'),
        ('p', 'OSt A MSG'),
        ('r', 'EAd All MSGS'),
        ('g', 'OOdbYE /lOGOff'),]

    # add LORD to menu only if enabled,
    if ini.CFG.getboolean('dosemu', 'enabled') and (
            ini.CFG.get('dosemu', 'lord_path') != 'no'):
        entries.insert(0, ('#', 'PlAY lORd!'))

    if 'sysop' in session.user.groups:
        entries += (('v', 'idEO CASSEttE'),)
    buf_str = u''
    for key, name in entries:
        out_str = u''.join((
            term.bold(u'('),
            term.bold_blue_underline(key),
            term.bold(u')'),
            term.bold_blue(name.split()[0]),
            u' ', u' '.join(name.split()[1:]),
            u'  '))
        ansilen = len(Ansi(buf_str + out_str))
        if ansilen >= (term.width * .8):
            echo(Ansi(buf_str).center(term.width) + u'\r\n\r\n')
            buf_str = out_str
        else:
            buf_str += out_str
    echo(Ansi(buf_str).center(term.width) + u'\r\n\r\n')
    echo(u' [%s]:' % (
        term.blue_underline(''.join([key for key, name in entries]))))


def main():
    """ Main procedure. """
    # pylint: disable=R0912
    #         Too many branches
    from x84.bbs import getsession, getch, goto, gosub
    session = getsession()

    inp = -1
    dirty = True
    while True:
        if dirty or session.poll_event('refresh'):
            refresh()
        inp = getch(1)
        dirty = True
        if inp == u'*':
            goto('main')  # reload main menu using hidden option '*'
        elif inp == u'b':
            gosub('bbslist')
        elif inp == u'l':
            gosub('lc')
        elif inp == u'o':
            gosub('ol')
        elif inp == u's':
            gosub('si')
        elif inp == u'w':
            gosub('online')
        elif inp == u'n':
            gosub('news')
        elif inp == u'f':
            gosub('weather')
        elif inp == u'e':
            gosub('profile')
        elif inp == u'#':
            gosub('lord')
        elif inp == u't':
            gosub('tetris')
        elif inp == u'c':
            gosub('chat')
        elif inp == u'p':
            gosub('writemsg')
        elif inp == u'r':
            gosub('readmsgs')
        elif inp == u'g':
            goto('logoff')
        elif inp == u'!':
            gosub('charset')
        elif inp == '\x1f' and 'sysop' in session.user.groups:
            # ctrl+_, run a debug script
            gosub('debug')
        elif inp == u'v' and 'sysop' in session.user.groups:
            # video cassette player
            gosub('ttyplay')
        else:
            dirty = False



"""
lightbar package for x/84 BBS, http://github.com/jquast/x84
"""
from x84.bbs.ansiwin import AnsiWindow

NETHACK_KEYSET = {'home': [u'y', '0'],
                  'end': [u'n', 'G'],
                  'pgup': [u'h', u'b'],
                  'pgdown': [u'l', u'f'],
                  'up': [u'k'],
                  'down': [u'j'],
                  'enter': [u'\r'],
                  'exit': [u'q', u'Q', unichr(27), ],
                  }


class Lightbar (AnsiWindow):
    """
    This Windowing class offers a classic 'lightbar' interface.

    Instantiate with yloc, xloc, height, and width, then call the update method
    with a list of unicode strings. send keycodes to process_keystroke () to
    interactive with the 'lightbar'.
    """
    # pylint: disable=R0902
    #         Too many instance attributes (15/7)
    # pylint: disable=R0904
    #         Too many public methods (29/20)
    content = list()

    def __init__(self, height, width, yloc, xloc):
        """
        Initialize a lightbar of height, width, y and x position.
        """
        AnsiWindow.__init__(self, height, width, yloc, xloc)
        self._vitem_idx = 0
        self._vitem_shift = 0
        self._vitem_lastidx = 0
        self._vitem_lastshift = 0
        self._selected = False
        self._quit = False
        self.keyset = NETHACK_KEYSET
        self.init_keystrokes()
        self.init_theme()

    def update(self, keyed_uchars=None):
        """
        Replace content of lightbar with iterable of aribitrary (key, unicode).
        unicode is displayed to the user, and key can be used for any
        programming purposes, such as sorting or identifying.
        """
        if keyed_uchars is None:
            keyed_uchars = (None, u'',)
        self.content = list(keyed_uchars)
        self.position = (self.vitem_idx, self.vitem_shift)

    def refresh_row(self, row):
        """
        Return unicode byte sequence suitable for moving to location ypos of
        window-relative row, and displaying any valid entry there, or using
        glyphs['erase'] if out of bounds. Strings are ansi color safe, and
        will be trimmed using glyphs['strip'] if their displayed width is
        wider than window.
        """
        import x84.bbs.session
        from x84.bbs.output import Ansi
        pos = self.pos(self.ypadding + row, self.xpadding)
        entry = self.vitem_shift + row
        if entry >= len(self.content):
            # out-of-bounds;
            return u''.join((pos,
                self.glyphs.get('erase', u' ') * self.visible_width,))

        def fit_row(ucs):
            """ Strip a unicode row to fit window boundry, if necessary """
            column = self.visible_width + 1
            wrapped = Ansi(ucs).wrap(column).splitlines()
            if len(wrapped) > 1:
                marker = self.glyphs.get('strip', u' $')
                marker_column = self.visible_width - len(marker)
                wrapped = Ansi(ucs).wrap(marker_column).splitlines()
                ucs = Ansi(wrapped[0].rstrip()).ljust(marker_column) + marker
                return ucs
            return (Ansi(ucs).ljust(column))

        term = x84.bbs.session.getterminal()
        # allow ucs data with '\r\n', to accomidate soft and hardbreaks; just
        # don't display them, really wrecks up cusor positioning.
        ucs = self.content[entry][1].strip(u'\r\n')

        # highlighted entry; strip of ansi sequences, use color 'highlight'
        # trim and append '$ ' if it cannot fit,
        if entry == self.index:
            ucs = Ansi(ucs).seqfill()
            if len(Ansi(ucs)) > self.visible_width:
                ucs = fit_row(ucs)
            return u''.join((pos,
                             self.colors.get('highlight', u''),
                             self.align(ucs),
                             term.normal,))
        # unselected entry; retain ansi sequences, decode any pipe characters,
        # trim and append '$ ' if it cannot fit
        ucs = Ansi(ucs).decode_pipe()
        if len(Ansi(ucs)) > self.visible_width:
            ucs = fit_row(ucs)
        return u''.join((pos,
                         self.colors.get('lowlight', u''),
                         self.align(ucs),
                         term.normal,))

    def fixate(self):
        """
        Reterm unicode terminal sequence for moving cursor to current
        selection.
        """
        return self.pos(self.ypadding + self.vitem_idx,
                        self.xpadding + self.visible_width)

    def refresh(self):
        """
        Refresh full lightbar window contents
        """
        return u''.join(self.refresh_row(ypos) for ypos in
                        range(max(self.visible_bottom, self.visible_height)))

    def refresh_quick(self):
        """
        Redraw only the 'dirty' portions after a 'move' has occured;
        otherwise redraw entire contents (page has shifted).
        """
        if self.moved:
            if (self._vitem_lastshift != self.vitem_shift):
                # page shift, refresh entire page
                return self.refresh()
            if self._vitem_lastidx != self.vitem_idx:
                # unhighlight last selection, highlight new
                return (self.refresh_row(self._vitem_lastidx)
                        + self.refresh_row(self.vitem_idx))
            else:
                # just highlight new ..
                return (self.refresh_row(self.vitem_idx))
        return u''

    def init_theme(self):
        """
        Initialize color['highlight'].
        """
        from x84.bbs.session import getterminal
        self.colors['highlight'] = getterminal().reverse_green
        self.glyphs['strip'] = u' $'  # indicates content was stripped
        AnsiWindow.init_theme(self)

    def init_keystrokes(self):
        """
        This initializer sets glyphs and colors appropriate for a "theme",
        override or inherit this method to create a common color and graphic
        set.
        """
        import x84.bbs.session
        term = x84.bbs.session.getterminal()
        self.keyset = NETHACK_KEYSET
        self.keyset['home'].append(term.KEY_HOME)
        self.keyset['end'].append(term.KEY_END)
        self.keyset['pgup'].append(term.KEY_PPAGE)
        self.keyset['pgdown'].append(term.KEY_NPAGE)
        self.keyset['up'].append(term.KEY_UP)
        self.keyset['down'].append(term.KEY_DOWN)
        self.keyset['enter'].append(term.KEY_ENTER)
        self.keyset['exit'].append(term.KEY_EXIT)

    def process_keystroke(self, key):
        """
        Process the keystroke received by run method and take action.
        """
        self._moved = False
        self._selected = False
        self._vitem_lastidx = self.vitem_idx
        self._vitem_lastshift = self.vitem_shift
        rstr = u''
        if key in self.keyset['home']:
            rstr = self.move_home()
        elif key in self.keyset['end']:
            rstr = self.move_end()
        elif key in self.keyset['pgup']:
            rstr = self.move_pageup()
        elif key in self.keyset['pgdown']:
            rstr = self.move_pagedown()
        elif key in self.keyset['up']:
            rstr = self.move_up()
        elif key in self.keyset['down']:
            rstr = self.move_down()
        elif key in self.keyset['enter']:
            self.selected = True
        elif key in self.keyset['exit']:
            self._quit = True
        return rstr

    def read(self):
        """
        Reads input until the ENTER or ESCAPE key is pressed (Blocking).
        Allows backspacing. Returns unicode text, or None when cancelled.
        """
        from x84.bbs import getch
        from x84.bbs.output import echo
        self._selected = False
        self._quit = False
        echo(self.refresh())
        while not (self.selected or self.quit):
            echo(self.process_keystroke(getch()))
        if self.quit:
            return None
        return self.selection[0]

    @property
    def quit(self):
        """
        Returns: True if a terminating or quit character was handled by
        process_keystroke(), such as the escape key, or 'q' by default.
        """
        return self._quit

    @property
    def index(self):
        """
        Selected index of self.content
        """
        return self.vitem_shift + self.vitem_idx

    @property
    def at_bottom(self):
        """
        Returns True if current selection is last in list
        """
        return self.index == len(self.content) - 1

    @property
    def at_top(self):
        """
        Returns True if current selection is first in list
        """
        return self.index == 0

    @property
    def selection(self):
        """
        Selected content of self.content by index
        """
        return (self.content[self.index]
                if self.index >= 0 and self.index < len(self.content)
                else (None, None))

    @property
    def selected(self):
        """
        Returns True when keyset['enter'] key detected in process_keystroke
        """
        return self._selected

    @selected.setter
    def selected(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        # this setter should only be used to reset to 'False' for recycling
        assert type(value) is bool
        self._selected = value

    @property
    def last_index(self):
        """
        Previously selected index of self.content
        """
        return self._vitem_lastshift + self._vitem_lastidx

    @property
    def position(self):
        """
        Tuple pair (row, page). 'row' is index from top of window,
        and 'page' is number of page items scrolled.
        """
        return (self.vitem_idx, self.vitem_shift)

    @position.setter
    def position(self, pos_tuple):
        # pylint: disable=C0111
        #         Missing docstring
        self.vitem_idx, self.vitem_shift = pos_tuple
        self._chk_bounds()

    @property
    def visible_content(self):
        """
        Returns content that is visible in window
        """
        # pylint: disable=W0612
        #        Unused variable 'item'
        item, shift = self.position
        return self.content[shift:shift + self.visible_height]

    @property
    def visible_bottom(self):
        """
        Visible bottom-most item of lightbar.
        """
        if self.vitem_shift + (self.visible_height - 1) > len(self.content):
            return len(self.content)
        else:
            return self.visible_height

    @property
    def vitem_idx(self):
        """
        Index of selected item relative by index to only the length of the list
        that is visible, without accounting for scrolled content.
        """
        # pylint: disable=C0111
        #         Missing docstring
        return self._vitem_idx

    @vitem_idx.setter
    def vitem_idx(self, value):
        # pylint: disable=C0111
        #        Missing docstring
        if self._vitem_idx != value:
            self._vitem_lastidx = self._vitem_idx
            self._vitem_idx = value
            self._moved = True

    @property
    def vitem_shift(self):
        """
        Index of top-most item in viewable window, non-zero when scrolled.
        This value effectively represents the number of items not in view
        due to paging.
        """
        # pylint: disable=C0111
        #         Missing docstring
        return self._vitem_shift

    @vitem_shift.setter
    def vitem_shift(self, value):
        # pylint: disable=C0111
        #        Missing docstring
        if self._vitem_shift != value:
            self._vitem_lastshift = self._vitem_shift
            self._vitem_shift = value
            self._moved = True

    def _chk_bounds(self):
        """
        Shift pages and selection until a selection is within bounds
        """
        # if selected item is out of range of new list, then scroll to last
        # page, and move selection to end of screen,
        if self.vitem_shift and (self.index + 1) > len(self.content):
            self.vitem_shift = len(self.content) - self.visible_height + 1
            self.vitem_idx = self.visible_height - 2

        # if we are a shifted window, shift 1 line up while keeping our
        # lightbar position until the bottom-most item is within visable range.
        while (self.vitem_shift and self.vitem_shift + self.visible_height - 1
                > len(self.content)):
            self.vitem_shift -= 1
            self.vitem_idx += 1

        # When a window is not shiftable, ensure selection is less than
        # total items. (truncate to last item)
        while self.vitem_idx > 0 and self.index >= len(self.content):
            self.vitem_idx -= 1

    def move_down(self):
        """
        Move selection down one row.
        """
        if self.at_bottom:
            # bounds check
            return u''
        if self.vitem_idx + 1 < self.visible_bottom:
            # move down 1 row
            self.vitem_idx += 1
        elif self.vitem_idx < len(self.content):
            # scroll down 1 row
            self.vitem_shift += 1
        return self.refresh_quick()

    def goto(self, index):
        """
        Move selection to index of lightbar content.
        """
        assert index >= 0 and index < len(self.content)
        row, shift = self.position
        while (row + shift) < index:
            if row < (self.visible_height - 1):
                row += 1
            else:
                shift += 1
        while (row + shift) > index:
            if row > 0:
                row -= 1
            else:
                shift -= 1
        self.position = row, shift
        return self.refresh_quick()

    def move_up(self):
        """
        Move selection up one row.
        """
        if self.at_top:
            # bounds check
            return u''
        elif self.vitem_idx >= 1:
            # move up 1 row
            self.vitem_idx -= 1
        elif self.vitem_shift > 0:
            # scroll up 1 row
            self.vitem_shift -= 1
        return self.refresh_quick()

    def move_pagedown(self):
        """
        Move selection down one page.
        """
        if len(self.content) < self.visible_height:
            # move to last entry
            if self.vitem_idx == len(self.content) - 1:
                return u''  # already at end
            self.vitem_idx = len(self.content) - 1
        elif (self.vitem_shift + self.visible_height
                < (len(self.content) - self.visible_height)):
            # previous page
            self.vitem_shift = self.vitem_shift + self.visible_height
        elif self.vitem_shift != len(self.content) - self.visible_height:
            # shift window to last page
            self.vitem_shift = len(self.content) - self.visible_height
        else:
            # already at last page, goto end
            return self.move_end()
        return self.refresh_quick()

    def move_pageup(self):
        """
        Move selection up one page.
        """
        if len(self.content) < self.visible_height - 1:
            self.vitem_idx = 0
        if self.vitem_shift - self.visible_height > 0:
            # previous page
            self.vitem_shift = self.vitem_shift - self.visible_height
        elif self.vitem_shift > 0:
            # shift window to first page
            self.vitem_shift = 0
        else:
            # already at first page, goto home
            return self.move_home()
        return self.refresh_quick()

    def move_home(self):
        """
        Move selection to the very top and first entry of the list.
        """
        if (0, 0) == (self.vitem_idx, self.vitem_shift):
            return u''  # already at home
        self.vitem_idx = 0
        self.vitem_shift = 0
        return self.refresh_quick()

    def move_end(self):
        """
        Move selection to the very last and final entry of the list.
        """
        if len(self.content) < self.visible_height:
            if self.vitem_idx == len(self.content) - 1:
                return u''  # already at end
            self.vitem_idx = len(self.content) - 1
        else:
            self.vitem_shift = len(self.content) - self.visible_height
            self.vitem_idx = self.visible_height - 1
        return self.refresh_quick()



#!/usr/bin/env python

# measure and report productivity in sense
# of work-weeks and work-hours, productive
# hours, and rest hours.

# ... a reference to the movie 'Visioneers'
msg_rest = u'There are %s minutes of rest remaining until work begins.'
msg_productivity = u'There are %s minutes of productivity remaining ' \
                   u'this work-week.'
from datetime import *

# productive times by week and starting & ending hour
p_time = \
  {0: {'h_start':9, 'h_end':18},
   1: {'h_start':9, 'h_end':18},
   2: {'h_start':9, 'h_end':18},
   3: {'h_start':9, 'h_end':18},
   4: {'h_start':9, 'h_end':18}}

tgt = now = datetime.now()

# target count down is always end of nearest work week
lastday = max(p_time)
lasthour = p_time[lastday]['h_end']

def productive(date):
  dt = date.timetuple()
  return False if not dt.tm_wday in p_time \
    or (dt.tm_hour < p_time[dt.tm_wday]['h_start'] or \
        dt.tm_hour >= p_time[dt.tm_wday]['h_end']) \
  else \
    True

def rest(date):
  minutes = 0
  tgt = date
  dt = tgt.timetuple()
  while not dt.tm_wday in p_time \
  or (not dt.tm_hour == p_time[dt.tm_wday]['h_start']):
    tgt += timedelta(minutes=1)
    dt = tgt.timetuple()
    minutes += 1
  return minutes

def remaining(date):
  tgt = date
  minutes = 0
  while not (tgt.timetuple().tm_wday == lastday \
  and tgt.timetuple().tm_hour == lasthour):
    tgt += timedelta(minutes=1)
    if productive(tgt):
      minutes += 1
  return minutes

def main():
    from x84.bbs import echo
    now = datetime.now()
    echo(u'\r\n\r\n')
    if not productive(now):
      echo(msg_rest % (rest(now),))
      echo(u'\r\n')
    echo(msg_productivity % (remaining(now),))
    echo(u'\r\n')

if __name__ == '__main__':
  now = datetime.now()
  if not productive(now):
    print(msg_rest % (rest(now),))
  print(msg_productivity % (remaining(now),))



""" message reader for x/84, https://github.com/jquast/x84 """
FILTER_PRIVATE = True
ALREADY_READ = set()
DELETED = set()
SEARCH_TAGS = set()
READING = False
TIME_FMT = '%A %b-%d, %Y at %r'


def quote_body(msg, width=79, quote_txt=u'> ', hardwrap=u'\r\n'):
    """
    Given a message, return new string suitable for quoting it.
    """
    from x84.bbs import Ansi
    ucs = u''
    for line in msg.body.splitlines():
        ucs += u''.join((
            quote_txt,
            Ansi(line).wrap(width - len(quote_txt), indent=quote_txt),
            hardwrap,))
    return u''.join((
        'On ',
        msg.stime.strftime(TIME_FMT), u' ',
        msg.author, ' wrote:',
        hardwrap, ucs, hardwrap))


def allow_tag(idx):
    """
    Returns true if user is allowed to 't'ag message at idx:
        * sysop and moderator
        * author or recipient
        * a member of any message tag matching user group
    """
    from x84.bbs import getsession, get_msg
    session = getsession()
    if ('sysop' in session.user.groups
            or 'moderator' in session.user.groups):
        return True
    msg = get_msg(idx)
    if session.user.handle in (msg.recipient, msg.author):
        return True
    for tag in msg.tags:
        if tag in session.user.groups:
            return True
    return False


def mark_undelete(idx):
    """ Mark message ``idx`` as deleted. """
    from x84.bbs import getsession
    session = getsession()
    # pylint: disable=W0603
    #         Using the global statement
    global DELETED
    DELETED = session.user.get('trash', set())
    if idx in DELETED:
        DELETED.remove(idx)
        session.user['trash'] = DELETED
        return True


def mark_delete(idx):
    """ Mark message ``idx`` as deleted. """
    from x84.bbs import getsession
    session = getsession()
    # pylint: disable=W0603
    #         Using the global statement
    global DELETED
    DELETED = session.user.get('trash', set())
    if not idx in DELETED:
        DELETED.add(idx)
        session.user['trash'] = DELETED
        return True


def mark_read(idx):
    """ Mark message ``idx`` as read. """
    from x84.bbs import getsession
    session = getsession()
    # pylint: disable=W0603
    #         Using the global statement
    global ALREADY_READ
    ALREADY_READ = session.user.get('readmsgs', set())
    if idx not in ALREADY_READ:
        ALREADY_READ.add(idx)
        session.user['readmsgs'] = ALREADY_READ
        return True


def msg_filter(msgs):
    """
        filter all matching messages. userland implementation
        of private/public messaging by using the 'tags' database.
        'new', or unread messages are marked by idx matching
        session.user['readmsgs'] when read. Finally, implement 'group'
        tagging, so that users of group 'impure' are allowed to read
        messages tagged by 'impure', regardless of recipient or 'public'.

        returns (msgs), (new). new contains redundant msgs
    """
    # pylint: disable=R0914,R0912,R0915
    #         Too many local variables
    #         Too many branches
    #         Too many statements
    from x84.bbs import list_msgs, echo, getsession, getterminal, get_msg, Ansi
    session, term = getsession(), getterminal()
    public_msgs = list_msgs(('public',))
    addressed_to = 0
    addressed_grp = 0
    filtered = 0
    deleted = 0
    private = 0
    public = 0
    new = set()
    echo(u' Processing ' + term.reverse_yellow('..'))
    for msg_id in msgs.copy():
        if msg_id in public_msgs:
            # can always ready msgs tagged with 'public'
            public += 1
        else:
            private += 1
        msg = get_msg(msg_id)
        if msg.recipient == session.user.handle:
            addressed_to += 1
        else:
            # a system of my own, by creating groups
            # with the same as tagged messages, you may
            # create private or intra-group messages.
            tag_matches_group = False
            for tag in msg.tags:
                if tag in session.user.groups:
                    tag_matches_group = True
                    break
            if tag_matches_group:
                addressed_grp += 1
            elif msg_id not in public_msgs:
                # denied to read this message
                if FILTER_PRIVATE:
                    msgs.remove(msg_id)
                    filtered += 1
                    continue
            elif msg_id in DELETED:
                msgs.remove(msg_id)
                deleted += 1
        if msg_id not in ALREADY_READ:
            new.add(msg_id)

    if 0 == len(msgs):
        echo(u'\r\n\r\nNo messages (%s filtered).' % (filtered,))
    else:
        txt_out = list()
        if addressed_to > 0:
            txt_out.append('%s addressed to you' % (
                term.bold_yellow(str(addressed_to)),))
        if addressed_grp > 0:
            txt_out.append('%s addressed by group' % (
                term.bold_yellow(str(addressed_grp)),))
        if filtered > 0:
            txt_out.append('%s filtered' % (
                term.bold_yellow(str(filtered)),))
        if deleted > 0:
            txt_out.append('%s deleted' % (
                term.bold_yellow(str(deleted)),))
        if public > 0:
            txt_out.append('%s public' % (
                term.bold_yellow(str(public)),))
        if private > 0:
            txt_out.append('%s private' % (
                term.bold_yellow(str(private)),))
        if len(new) > 0:
            txt_out.append('%s new' % (
                term.bold_yellow(str(len(new),)),))
        if 0 != len(txt_out):
            echo(u'\r\n\r\n' + Ansi(
                u', '.join(txt_out) + u'.').wrap(term.width - 2))
    return msgs, new


def banner():
    """ Returns string suitable for displaying banner. """
    from x84.bbs import getterminal
    term = getterminal()
    return u''.join((u'\r\n\r\n',
                     term.yellow(u'... '.center(term.width).rstrip()),
                     term.bold_yellow(' MSG R'),
                     term.yellow('EAdER'),))


def prompt_tags(tags):
    """ Prompt for and return valid tags from TAGDB. """
    # pylint: disable=R0914,W0603
    #         Too many local variables
    #         Using the global statement
    from x84.bbs import DBProxy, echo, getterminal, getsession
    from x84.bbs import Ansi, LineEditor, getch
    session, term = getsession(), getterminal()
    tagdb = DBProxy('tags')
    global FILTER_PRIVATE
    while True:
        # Accept user input for a 'search tag', or /list command
        #
        echo(u"\r\n\r\nENtER SEARCh %s, COMMA-dEliMitEd. " % (
            term.red('TAG(s)'),))
        echo(u"OR '/list', %s%s\r\n : " % (
            (term.yellow_underline('^a') + u':utoscan '
                ) if session.user.get('autoscan', False) else u'',
            term.yellow_underline('Escape') + u':quit',))
        width = term.width - 6
        sel_tags = u', '.join(tags)
        while len(Ansi(sel_tags)) >= (width - 8):
            tags = tags[:-1]
            sel_tags = u', '.join(tags)
        lne = LineEditor(width, sel_tags)
        while not lne.carriage_returned:
            inp = getch()
            if inp in (unichr(1),): # ^A:
                inp_tags = session.user.get('autoscan', set())
                break
            else:
                echo(lne.process_keystroke(inp))
        inp_tags = lne.content
        if (inp_tags is None or 0 == len(inp_tags)
                or inp_tags.strip().lower() == '/quit'):
            return None
        elif inp_tags.strip().lower() == '/list':
            # list all available tags, and number of messages
            echo(u'\r\n\r\nTags: \r\n')
            all_tags = sorted(tagdb.items())
            if 0 == len(all_tags):
                echo(u'None !'.center(term.width / 2))
            else:
                echo(Ansi(u', '.join(([u'%s(%s)' % (
                    term.red(tag),
                    term.yellow(str(len(msgs))),)
                        for (tag, msgs) in all_tags]))).wrap(term.width - 2))
            continue
        elif (inp_tags.strip().lower() == '/nofilter'
                and 'sysop' in session.user.groups):
            # disable filtering private messages
            FILTER_PRIVATE = False
            continue

        echo(u'\r\n')
        # search input as valid tag(s)
        tags = set([_tag.strip().lower() for _tag in inp_tags.split(',')])
        for tag in tags.copy():
            if not tag in tagdb:
                tags.remove(tag)
                echo(u"\r\nNO MESSAGES With tAG '%s' fOUNd." % (
                    term.red(tag),))
        return tags


def main(autoscan_tags=None):
    """ Main procedure. """
    # pylint: disable=W0603,R0912
    #         Using the global statement
    #         Too many branches
    from x84.bbs import getsession, getterminal, echo, getch
    from x84.bbs import list_msgs
    session, term = getsession(), getterminal()
    session.activity = 'autoscan msgs'
    echo(banner())
    global ALREADY_READ, SEARCH_TAGS, DELETED
    if autoscan_tags is not None:
        SEARCH_TAGS = autoscan_tags
        echo(u''.join((
            term.bold_black('[ '),
            term.yellow('AUtOSCAN'),
            term.bold_black(' ]'), u'\r\n')))
    else:
        SEARCH_TAGS = set(['public'])
        # also throw in user groups
        SEARCH_TAGS.update(session.user.groups)
        SEARCH_TAGS = prompt_tags(SEARCH_TAGS)
        if SEARCH_TAGS is None:
            SEARCH_TAGS = set()

    echo(u'\r\n\r\n%s%s ' % (
        term.bold_yellow('SCANNiNG'),
        term.bold_black(':'),))
    echo(u','.join([term.red(tag) for tag in SEARCH_TAGS]
                   if 0 != len(SEARCH_TAGS) else ['<All>', ]))

    if (SEARCH_TAGS != session.user.get('autoscan', None)):
        echo(u'\r\n\r\nSave tag list as autoscan on login [yn] ?\b\b')
        while True:
            inp = getch()
            if inp in (u'q', 'Q', unichr(27), u'n', u'N'):
                break
            elif inp in (u'y', u'Y'):
                session.user['autoscan'] = SEARCH_TAGS
                break

    # retrieve all matching messages,
    all_msgs = list_msgs(SEARCH_TAGS)
    echo(u'\r\n\r\n%s messages.' % (term.yellow_reverse(str(len(all_msgs),))))
    if 0 == len(all_msgs):
        getch(0.5)
        return

    # filter messages public/private/group-tag/new
    ALREADY_READ = session.user.get('readmsgs', set())
    DELETED = session.user.get('trash', set())
    msgs, new = msg_filter(all_msgs)
    if 0 == len(msgs) and 0 == len(new):
        getch(0.5)
        return

    # prompt read 'a'll, 'n'ew, or 'q'uit
    echo(u'\r\n  REAd [%s]ll %d%s message%s [qa%s] ?\b\b' % (
        term.yellow_underline(u'a'),
        len(msgs), (
        u' or %d [%s]EW ' % (
        len(new), term.yellow_underline(u'n'),)
            if new else u''),
        u's' if 1 != len(msgs) else u'',
        u'n' if new else u'',))
    while True:
        inp = getch()
        if inp in (u'q', 'Q', unichr(27)):
            return
        elif inp in (u'n', u'N') and len(new):
            # read only new messages
            msgs = new
            break
        elif inp in (u'a', u'A'):
            break

    # read target messages
    read_messages(msgs, new)


def read_messages(msgs, new):
    """
    Provide message reader UI given message list ``msgs``,
    with new messages in list ``new``.
    """
    # pylint: disable=R0914,R0912,R0915
    #         Too many local variables
    #         Too many branches
    #         Too many statements
    from x84.bbs import timeago, get_msg, getterminal, echo, gosub
    from x84.bbs import ini, Pager, getsession, getch, Ansi, Msg
    import x84.default.writemsg
    session, term = getsession(), getterminal()

    session.activity = 'reading msgs'
    # build header
    len_idx = max([len('%d' % (_idx,)) for _idx in msgs])
    len_author = ini.CFG.getint('nua', 'max_user')
    len_ago = 9
    len_subject = ini.CFG.getint('msg', 'max_subject')
    len_preview = min(len_idx + len_author + len_ago + len_subject + 4,
                      term.width - 2)
    reply_depth = ini.CFG.getint('msg', 'max_depth')
    indent_start, indent, indent_end = u'\\', u'-', u'> '

    def get_header(msgs_idx):
        """
        Return list of tuples, (idx, unicodestring), suitable for Lightbar.
        """
        import datetime
        msg_list = list()
        thread_indent = lambda depth: (term.red(
            (indent_start + (indent * depth) + indent_end))
            if depth else u'')

        def head(msg, depth=0, maxdepth=reply_depth):
            """ This recursive routine finds the 'head' message
                of any relationship, up to maxdepth.
            """
            if (depth <= maxdepth
                    and hasattr(msg, 'parent')
                    and msg.parent is not None):
                return head(get_msg(msg.parent), depth + 1, maxdepth)
            return msg.idx, depth

        for idx in msgs_idx:
            msg = get_msg(idx)
            author, subj = msg.author, msg.subject
            tm_ago = (datetime.datetime.now() - msg.stime).total_seconds()
            # pylint: disable=W0631
            #         Using possibly undefined loop variable 'idx'
            attr = lambda arg: (
                term.bold_green(arg) if (
                    not idx in ALREADY_READ
                    and msg.recipient == session.user.handle) else
                term.red(arg) if not idx in ALREADY_READ
                else term.yellow(arg))
            status = [u'U' if not idx in ALREADY_READ else u' ',
                      u'D' if idx in DELETED else u' ', ]
            row_txt = u'%s %s %s %s %s%s ' % (
                u''.join(status),
                attr(str(idx).rjust(len_idx)),
                attr(author.ljust(len_author)),
                (timeago(tm_ago)).rjust(len_ago),
                attr(u'ago'),
                term.bold_black(':'),)
            msg_list.append((head(msg), idx, row_txt, subj))
        msg_list.sort()
        return [(idx, row_txt + thread_indent(depth) + subj)
                for (_threadid, depth), idx, row_txt, subj in msg_list]

    def get_selector(mailbox, prev_sel=None):
        """
        Provide Lightbar UI element given message mailbox returned from
        function get_header, and prev_sel as previously instantiated Lightbar.
        """
        from x84.bbs import Lightbar
        pos = prev_sel.position if prev_sel is not None else (0, 0)
        sel = Lightbar(
            height=(term.height / 3
                    if term.width < 140 else term.height - 3),
            width=len_preview,
            yloc=2, xloc=0)
        sel.glyphs['top-horiz'] = u''
        sel.glyphs['left-vert'] = u''
        sel.colors['highlight'] = term.yellow_reverse
        sel.update(mailbox)
        sel.position = pos
        return sel

    def get_reader():
        """
        Provide Pager UI element for message reading.
        """
        reader_height = (term.height - (term.height / 3) - 2)
        reader_indent = 2
        reader_width = min(term.width - 1, min(term.width - reader_indent, 80))
        reader_ypos = ((term.height - 1 ) - reader_height if
                      (term.width - reader_width) < len_preview else 2)
        reader_height = term.height - reader_ypos - 1
        msg_reader = Pager(
            height=reader_height,
            width=reader_width,
            yloc=reader_ypos,
            xloc=min(len_preview + 2, term.width - reader_width))
        msg_reader.glyphs['top-horiz'] = u''
        msg_reader.glyphs['right-vert'] = u''
        return msg_reader

    def format_msg(reader, idx):
        """ Format message of index ``idx`` into Pager instance ``reader``. """
        msg = get_msg(idx)
        sent = msg.stime.strftime(TIME_FMT)
        to_attr = term.bold_green if (
            msg.recipient == session.user.handle) else term.underline
        ucs = u'\r\n'.join((
            (u''.join((
                term.yellow('fROM: '),
                (u'%s' % term.bold(msg.author,)).rjust(len_author),
                u' ' * (reader.visible_width - (len_author + len(sent))),
                sent,))),
            u''.join((
                term.yellow('tO: '),
                to_attr((u'%s' % to_attr(msg.recipient,)).rjust(len_author)
                        if msg.recipient is not None else u'All'),)),
            (Ansi(
                term.yellow('tAGS: ')
                + (u'%s ' % (term.bold(','),)).join((
                    [term.bold_red(_tag)
                        if _tag in SEARCH_TAGS
                        else term.yellow(_tag)
                        for _tag in msg.tags]))).wrap(
                            reader.visible_width,
                            indent=u'      ')),
            (term.yellow_underline(
                (u'SUbj: %s' % (msg.subject,)).ljust(reader.visible_width)
            )),
            u'', (msg.body),))
        return ucs

    def get_selector_title(mbox, new):
        """
        Returns unicode string suitable for displaying as title of mailbox.
        """
        newmsg = (term.yellow(u' ]-[ ') +
                  term.yellow_reverse(str(len(new))) +
                  term.bold_underline(u' NEW')) if len(new) else u''
        return u''.join((term.yellow(u'[ '),
                         term.bold_yellow(str(len(mbox))),
                         term.bold(
                         u' MSG%s' % (u's' if 1 != len(mbox) else u'',)),
                         newmsg, term.yellow(u' ]'),))

    dispcmd_mark = lambda idx: (
        (term.yellow_underline(u' ') + u':mark' + u' ')
        if idx not in ALREADY_READ else u'')
    dispcmd_delete = lambda idx: (
        (term.yellow_underline(u'D') + u':elete' + u' ')
        if idx not in DELETED else u'')
    dispcmd_tag = lambda idx: (
        (term.yellow_underline(u't') + u':ag' + u' ')
        if allow_tag(idx) else u'')

    def get_selector_footer(idx):
        """
        Returns unicode string suitable for displaying
        as footer of mailbox when window is active.
        """
        return u''.join((
            term.yellow(u'- '),
            u''.join((
                term.yellow_underline(u'>') + u':read ',
                term.yellow_underline(u'r') + u':eply ',
                dispcmd_mark(idx),
                dispcmd_delete(idx),
                dispcmd_tag(idx),
                term.yellow_underline(u'q') + u':uit',)),
            term.yellow(u' -'),))

    def get_reader_footer(idx):
        """
        Returns unicode string suitable for displaying
        as footer of reader when window is active
        """

        return u''.join((
            term.yellow(u'- '),
            u' '.join((
                term.yellow_underline(u'<') + u':back ',
                term.yellow_underline(u'r') + u':eply ',
                dispcmd_delete(idx),
                dispcmd_tag(idx),
                term.yellow_underline(u'q') + u':uit',)),
            term.yellow(u' -'),))

    def refresh(reader, selector, mbox, new):
        """
        Returns unicode string suitable for refreshing the screen.
        """
        if READING:
            reader.colors['border'] = term.bold_yellow
            selector.colors['border'] = term.bold_black
        else:
            reader.colors['border'] = term.bold_black
            selector.colors['border'] = term.bold_yellow
        title = get_selector_title(mbox, new)
        padd_attr = (term.bold_yellow if not READING
                     else term.bold_black)
        sel_padd_right = padd_attr(
            u'-'
            + selector.glyphs['bot-horiz'] * (
            selector.visible_width - len(Ansi(title)) - 7)
            + u'-\u25a0-' if READING else u'- -')
        sel_padd_left = padd_attr(
            selector.glyphs['bot-horiz'] * 3)
        idx = selector.selection[0]
        return u''.join((term.move(0, 0), term.clear, u'\r\n',
                         u'// REAdiNG MSGS ..'.center(term.width).rstrip(),
                         selector.refresh(),
                         reader.border(),
                         selector.border(),
                         selector.title(
                             sel_padd_left + title + sel_padd_right),
                         selector.footer(get_selector_footer(idx)
                                         ) if not READING else u'',
                         reader.footer(get_reader_footer(idx)
                                       ) if READING else u'',
                         reader.refresh(),
                         ))

    echo((u'\r\n' + term.clear_eol) * (term.height - 1))
    dirty = 2
    msg_selector = None
    msg_reader = None
    idx = None
    # pylint: disable=W0603
    #         Using the global statement
    global READING
    while (msg_selector is None and msg_reader is None
           ) or not (msg_selector.quit or msg_reader.quit):
        if session.poll_event('refresh'):
            dirty = 2
        if dirty:
            if dirty == 2:
                mailbox = get_header(msgs)
            msg_selector = get_selector(mailbox, msg_selector)
            idx = msg_selector.selection[0]
            msg_reader = get_reader()
            msg_reader.update(format_msg(msg_reader, idx))
            echo(refresh(msg_reader, msg_selector, msgs, new))
            dirty = 0
        inp = getch(1)
        if inp in (u'r', u'R'):
            reply_to = get_msg(idx)
            reply_msg = Msg()
            reply_msg.recipient = reply_to.author
            reply_msg.tags = reply_to.tags
            reply_msg.subject = reply_to.subject
            reply_msg.parent = reply_to.idx
            # quote between 30 and 79, 'screen width - 4' as variable dist.
            reply_msg.body = quote_body(reply_to,
                                        max(30, min(79, term.width - 4)))
            echo(term.move(term.height, 0) + u'\r\n')
            if gosub('writemsg', reply_msg):
                reply_msg.save()
                dirty = 2
                READING = False
            else:
                dirty = 1
            mark_read(idx)  # also mark as read

        # 't' uses writemsg.prompt_tags() routine
        elif inp in (u't',) and allow_tag(idx):
            echo(term.move(term.height, 0))
            msg = get_msg(idx)
            if x84.default.writemsg.prompt_tags(msg):
                msg.save()
            dirty = 2

        # spacebar marks as read, goes to next message
        elif inp in (u' ',):
            dirty = 2 if mark_read(idx) else 1
            msg_selector.move_down()
            idx = msg_selector.selection[0]
            READING = False

        # D marks as deleted, goes to next message
        elif inp in (u'D',):
            dirty = 2 if mark_delete(idx) else 1
            msg_selector.move_down()
            idx = msg_selector.selection[0]
            READING = False

        # U undeletes, does not move.
        elif inp in (u'U',):
            dirty = 2 if mark_undelete(idx) else 1
            msg_selector.move_down()
            idx = msg_selector.selection[0]
            READING = False

        if READING:
            echo(msg_reader.process_keystroke(inp))
            # left, <, or backspace moves UI
            if inp in (term.KEY_LEFT, u'<', u'h',
                       '\b', term.KEY_BACKSPACE):
                READING = False
                dirty = 1
        else:
            echo(msg_selector.process_keystroke(inp))
            idx = msg_selector.selection[0]
            # right, >, or enter marks message read, moves UI
            if inp in (u'\r', term.KEY_ENTER, u'>',
                       u'l', 'L', term.KEY_RIGHT):
                dirty = 2 if mark_read(idx) else 1
                READING = True
            elif msg_selector.moved:
                dirty = 1
    echo(term.move(term.height, 0) + u'\r\n')
    return



"""
 Configuration for x/84 BBS, https://github.com/jquast/x84/
"""

import logging.config
import os.path

CFG = None

# pylint: disable=R0915,R0912,W0603
#         Too many statements
#         Too many branches
#         Using the global statement


def init(lookup_bbs, lookup_log):
    """
    Initialize global 'CFG' variable, a singleton to contain bbs properties and
    settings across all modules, as well as the logger. Each variable is tuple
    lookup path of in-order preferences for .ini files.

    If none our found, defaults are initialized, and the last item of each
    tuple is created.
    """
    import ConfigParser
    root = logging.getLogger()

    def write_cfg(cfg, filepath):
        """
        Write Config to filepath.
        """
        if not os.path.exists(os.path.dirname(os.path.expanduser(filepath))):
            print('Creating folder %s\n' % (os.path.dirname(os.path.expanduser(filepath)),))
            os.mkdir(os.path.dirname(os.path.expanduser(filepath)))
        print('Saving %s\n' % (filepath,))
        cfg.write(open(os.path.expanduser(filepath), 'wb'))

    # exploit last argument, presumed to be within a folder
    # writable by our process, and where the ini is wanted
    # -- engine.py specifys a default of: ~/.x84/somefile.ini
    loaded = False
    cfg_logfile = lookup_log[-1]
    for cfg_logfile in lookup_log:
        cfg_logfile = os.path.expanduser(cfg_logfile)
        # load-only defaults,
        if os.path.exists(cfg_logfile):
            print ('loading %s' % (cfg_logfile,))
            logging.config.fileConfig(cfg_logfile)
            loaded = True
            break
    if not loaded:
        cfg_log = init_log_ini()
        if not os.path.isdir(os.path.dirname(cfg_logfile)):
            try:
                os.mkdir(os.path.dirname(cfg_logfile))
            except OSError as err:
                root.warn('%s', err)
        try:
            write_cfg(cfg_log, cfg_logfile)
            root.info('Saved %s' % (cfg_logfile,))
        except IOError as err:
            root.error('%s', err)
        print(cfg_logfile)
        logging.config.fileConfig(cfg_logfile)

    loaded = False
    cfg_bbs = ConfigParser.SafeConfigParser()
    cfg_bbsfile = lookup_bbs[-1]
    for cfg_bbsfile in lookup_bbs:
        cfg_bbsfile = os.path.expanduser(cfg_bbsfile)
        # load defaults,
        if os.path.exists(cfg_bbsfile):
            cfg_bbs.read(cfg_bbsfile)
            root.info('loaded %s', cfg_bbsfile)
            loaded = True
            break
    if not loaded:
        cfg_bbs = init_bbs_ini()
        if not os.path.isdir(os.path.dirname(cfg_bbsfile)):
            try:
                os.mkdir(os.path.dirname(cfg_bbsfile))
            except OSError as err:
                root.warn('%s', err)
        try:
            write_cfg(cfg_bbs, cfg_bbsfile)
            root.info('Saved %s' % (cfg_bbsfile,))
        except IOError as err:
            root.error('%s', err)

    global CFG
    CFG = cfg_bbs


def init_bbs_ini():
    """
    Returns ConfigParser instance of bbs system defaults
    """
    import ConfigParser
    import getpass
    import socket
    cfg_bbs = ConfigParser.SafeConfigParser()

    cfg_bbs.add_section('system')
    cfg_bbs.set('system', 'bbsname', 'x/84')
    cfg_bbs.set('system', 'sysop', '')
    cfg_bbs.set('system', 'software', 'x/84')
    # use module-level 'default' folder
    cfg_bbs.set('system', 'scriptpath', os.path.abspath(
        os.path.join(os.path.dirname(__file__), os.path.pardir, 'default')))
    cfg_bbs.set('system', 'datapath',
                os.path.join(os.path.expanduser('~/.x84'), 'data'))
    cfg_bbs.set('system', 'ttyrecpath',
                os.path.join(os.path.expanduser('~/.x84'), 'ttyrecordings'))
    cfg_bbs.set('system', 'timeout', '1984')
    # for very slow systems, you may need to increase IPC timeout for acquiring
    # locks and sending input to sub-processes -- this can happen when the
    # system is under very heavy load -- like pasting wikipedia into the editor
    cfg_bbs.set('system', 'timeout_ipc', '1')
    cfg_bbs.set('system', 'password_digest', 'internal')
    cfg_bbs.set('system', 'mail_addr',
                '%s@%s' % (getpass.getuser(), socket.gethostname()))
    cfg_bbs.set('system', 'mail_smtphost', 'localhost')
    # change 'ansi' termcaps to 'ansi-bbs', for SynchoTerm
    cfg_bbs.set('system', 'termcap-ansi', 'ansi-bbs')
    # change 'unknown' termcaps to 'vt220', for dumb terminals
    cfg_bbs.set('system', 'termcap-unknown', 'vt220')
    # could be information leak to sensitive sysops
    cfg_bbs.set('system', 'show_traceback', 'no')
    # store passwords in uppercase, facebook and mystic bbs does this ..
    cfg_bbs.set('system', 'pass_ucase', 'no')

    cfg_bbs.add_section('telnet')
    cfg_bbs.set('telnet', 'addr', '127.0.0.1')
    cfg_bbs.set('telnet', 'port', '6023')

    # default path if cmd argument is not absolute,
    cfg_bbs.add_section('door')
    cfg_bbs.set('door', 'path', '/usr/local/bin:/usr/games')

    cfg_bbs.add_section('matrix')
    cfg_bbs.set('matrix', 'newcmds', 'new apply')
    cfg_bbs.set('matrix', 'byecmds', 'exit logoff bye quit')
    cfg_bbs.set('matrix', 'script', 'matrix')
    cfg_bbs.set('matrix', 'topscript', 'top')
    cfg_bbs.set('matrix', 'enable_anonymous', 'no')
    cfg_bbs.set('matrix', 'enable_pwreset', 'yes')

    cfg_bbs.add_section('session')
    cfg_bbs.set('session', 'record_tty', 'yes')
    cfg_bbs.set('session', 'tap_input', 'no')
    cfg_bbs.set('session', 'tap_output', 'no')
    cfg_bbs.set('session', 'tap_events', 'no')
    cfg_bbs.set('session', 'tap_db', 'no')
    cfg_bbs.set('session', 'default_encoding', 'utf8')

    cfg_bbs.add_section('irc')
    cfg_bbs.set('irc', 'server', 'efnet.xs4all.nl')
    cfg_bbs.set('irc', 'port', '6667')
    cfg_bbs.set('irc', 'channel', '#prsv')

    cfg_bbs.add_section('nethack')
    cfg_bbs.set('nethack', 'enabled', 'yes')
    cfg_bbs.set('nethack', 'path', '/nh343/nethack.343-nao')
    cfg_bbs.set('nethack', 'logfile', '/nh343/var/xlogfile')

    cfg_bbs.add_section('nua')
    cfg_bbs.set('nua', 'script', 'nua')
    cfg_bbs.set('nua', 'min_user', '3')
    cfg_bbs.set('nua', 'min_pass', '4')
    cfg_bbs.set('nua', 'max_user', '11')
    cfg_bbs.set('nua', 'max_pass', '16')
    cfg_bbs.set('nua', 'max_email', '50')
    cfg_bbs.set('nua', 'max_location', '24')
    cfg_bbs.set('nua', 'allow_apply', 'yes')
    cfg_bbs.set('nua', 'invalid_handles', ' '.join(
        (cfg_bbs.get('matrix', 'byecmds'),
         cfg_bbs.get('matrix', 'newcmds'),
         'sysop anonymous',)))
    cfg_bbs.set('nua', 'handle_validation', '^[A-Za-z0-9]{3,11}$')

    cfg_bbs.add_section('msg')
    cfg_bbs.set('msg', 'max_subject', '40')
    # maximum recursion for searching 'head' in a reply-to chain;
    # as each get_msg() is a lookup, thread-related sorting could
    # become too expensive.
    cfg_bbs.set('msg', 'max_depth', '8')

    cfg_bbs.add_section('dosemu')
    cfg_bbs.set('dosemu', 'enabled', 'no')
    cfg_bbs.set('dosemu', 'bin', '/usr/bin/dosemu')
    cfg_bbs.set('dosemu', 'home', '/DOS')
    # set to a valid folder to enable lord; dropfile is placed in lord folder.
    cfg_bbs.set('dosemu', 'lord_path', '/DOS/X/lord')
    cfg_bbs.set('dosemu', 'lord_dropfile', 'DORINFO')
    cfg_bbs.set('dosemu', 'lord_args',
            '-quiet -I \'$_com1 = "virtual"\' \'X:\\LORD\\START.BAT %%#\'')

    cfg_bbs.add_section('ttyplay')
    cfg_bbs.set('ttyplay', 'exe', '/usr/bin/ttyplay')
    return cfg_bbs

def init_log_ini():
    """
    Returns ConfigParser instance of logger defaults
    """
    import ConfigParser
    cfg_log = ConfigParser.RawConfigParser()
    cfg_log.add_section('formatters')
    cfg_log.set('formatters', 'keys', 'default')

    cfg_log.add_section('formatter_default')
    # for multiprocessing/threads, use: %(processName)s %(threadName) !
    cfg_log.set('formatter_default', 'format',
                u'%(asctime)s %(levelname)s '
                u'%(filename)11s:%(lineno)-3s %(message)s')
    cfg_log.set('formatter_default', 'class', 'logging.Formatter')
    cfg_log.set('formatter_default', 'datefmt', '%a-%m-%d %I:%M%p')

    cfg_log.add_section('handlers')
    cfg_log.set('handlers', 'keys', 'console, rotate_daily')

    cfg_log.add_section('handler_console')
    cfg_log.set('handler_console', 'class',
                'x84.bbs.log.ColoredConsoleHandler')
    cfg_log.set('handler_console', 'formatter', 'default')
    cfg_log.set('handler_console', 'args', 'tuple()')

    cfg_log.add_section('handler_rotate_daily')
    cfg_log.set('handler_rotate_daily', 'class',
                'logging.handlers.TimedRotatingFileHandler')
    cfg_log.set('handler_rotate_daily', 'level', 'INFO')
    cfg_log.set('handler_rotate_daily', 'suffix', '%Y%m%d')
    cfg_log.set('handler_rotate_daily', 'encoding', 'utf8')
    cfg_log.set('handler_rotate_daily', 'formatter', 'default')
    daily_log = os.path.join(os.path.expanduser('~/.x84'), 'daily.log')
    cfg_log.set('handler_rotate_daily', 'args',
                '("' + daily_log + '", "midnight", 1, 60)')

    cfg_log.add_section('loggers')
    cfg_log.set('loggers', 'keys', 'root, sqlitedict')

    cfg_log.add_section('logger_root')
    cfg_log.set('logger_root', 'level', 'INFO')
    cfg_log.set('logger_root', 'formatter', 'default')
    cfg_log.set('logger_root', 'handlers', 'console, rotate_daily')

    # squelche sqlitedict's info on open, its rather long
    cfg_log.add_section('logger_sqlitedict')
    cfg_log.set('logger_sqlitedict', 'level', 'WARN')
    cfg_log.set('logger_sqlitedict', 'formatter', 'default')
    cfg_log.set('logger_sqlitedict', 'handlers', 'console, rotate_daily')
    cfg_log.set('logger_sqlitedict', 'qualname', 'sqlitedict')
    cfg_log.set('logger_sqlitedict', 'propagate', '0')

    return cfg_log



"""
Weather retriever for x/84 https://github.com/jquast/x84
"""
from xml.etree import cElementTree as ET
import requests

# sorry this isn't doing celcius, etc., its just how the data comes,
# conversions were done in previous versions, but removed when rewritten
# for brevity.

def disp_msg(msg):
    """ Display unicode string ``msg`` in yellow. """
    from x84.bbs import getterminal, echo
    term = getterminal()
    echo(u''.join((u'\r\n\r\n',
                   term.bold_yellow('%s ' % (msg,),),
                   term.yellow_reverse_bold(u'...'),)))


def disp_notfound():
    """ Display 'bad request -/- not found in red. """
    from x84.bbs import getsession, getterminal, echo, getch
    term = getterminal()
    echo(u''.join((u'\r\n\r\n',
                   term.bold(u'bAd REQUESt'),
                   term.bold_red(' -/- '),
                   term.bold('NOt fOUNd.',),)))
    if not getsession().user.get('expert', False):
        getch(1.7)


def disp_found(num):
    """ Display 'N locations discovered' in yellow/white. """
    from x84.bbs import getterminal, echo
    term = getterminal()
    echo(u''.join((u'\r',
                   term.bold_white(u'%d' % (num,)),
                   term.yellow(u' lOCAtiON%s diSCOVEREd ' %
                               (u's' if num > 1 else u'')),
                   term.bold_black(u'...'),)))


def disp_search_help():
    """ Display searchbar usage. """
    from x84.bbs import getterminal, echo
    term = getterminal()
    msg_enterzip = (
        term.yellow(u'ENtER US '),
        term.bold_yellow(u'POStAl COdE'),
        term.yellow(u', OR NEARESt '),
        term.bold_yellow(u'iNtERNAtiONAl CitY. '),
        term.bold_yellow(u'('),
        term.underline_yellow('Escape'),
        term.bold_white(u':'),
        term.yellow('EXit'),
        term.bold_yellow(u')'),)
    echo(u''.join((u'\r\n\r\n', term.normal,) + msg_enterzip))


def do_fetch(postal):
    """
    Given postal code, fetch and return xml root node of weather results.
    """
    from x84.bbs import echo, getch, getterminal
    import StringIO
    term = getterminal()
    disp_msg('fEtChiNG')
    resp = requests.get(u'http://apple.accuweather.com'
                        + u'/adcbin/apple/Apple_Weather_Data.asp',
                        params=(('zipcode', postal),))
    if resp is None:
        disp_notfound()
        return None
    if resp.status_code != 200:
        # todo: logger.error
        echo(u'\r\n')
        echo(term.bold_red(u'StAtUS COdE: %s' % (resp.status_code,)))
        echo(u'\r\n\r\n')
        echo(repr(resp.content))
        echo(u'\r\n\r\n' + 'PRESS ANY kEY')
        getch()
        return None
    print resp.content
    xml_stream = StringIO.StringIO(resp.content)
    tree = ET.parse(xml_stream)
    return tree.getroot()


def do_search(search):
    """
    Given any arbitrary string, return list of possible matching locations.
    """
    import StringIO
    from x84.bbs import echo, getch
    disp_msg('SEARChiNG')
    resp = requests.get(u'http://apple.accuweather.com'
                        + u'/adcbin/apple/Apple_find_city.asp',
                        params=(('location', search),))
    locations = list()
    if resp is None:
        disp_notfound()
    elif resp.status_code != 200:
        # todo: logger.error
        echo(u'\r\n' + u'StAtUS COdE: %s\r\n\r\n' % (resp.status_code,))
        echo(repr(resp.content))
        echo(u'\r\n\r\n' + 'PRESS ANY kEY')
        getch()
    else:
        print resp.content
        xml_stream = StringIO.StringIO(resp.content)
        locations = list([dict(elem.attrib.items())
                          for _event, elem in ET.iterparse(xml_stream)
                          if elem.tag == 'location'])
        if 0 == len(locations):
            disp_notfound()
        else:
            disp_found(len(locations))
    return locations


def parse_weather(root):
    """
    Parse and return dictionary describing today's weather
    from weather xml root node.
    """
    weather = dict()
    # parse all current conditions from XML, value is cdata.
    for elem in root.find('CurrentConditions'):
        weather[elem.tag] = elem.text.strip() if elem.text is not None else u''
        # store attribute values
        for attr, val in elem.attrib.items():
            weather['%s-%s' % (elem.tag, attr)] = val
    return weather


def parse_forecast(root):
    """
    Parse and return dictionary describing weather forecast
    from weather xml root node.
    """
    forecast = dict()
    for elem in root.find('Forecast'):
        if elem.tag == 'day':
            key = elem.attrib.get('number')
            forecast[key] = dict()
            for subelem in elem:
                forecast[key][subelem.tag] = subelem.text.strip()
    return forecast


def get_zipsearch(zipcode=u''):
    """
    Prompt user for zipcode or international city.
    """
    from x84.bbs import getterminal, LineEditor, echo
    term = getterminal()
    echo(u''.join((u'\r\n\r\n',
                   term.bold_yellow(u'  -'),
                   term.reverse_yellow(u':'),
                   u' ')))
    return LineEditor(width=min(30, term.width - 5), content=zipcode).read()


def chose_location_dummy(locations):
    """
    dummy pager for chosing a location
    """
    from x84.bbs import getterminal, echo, getch, LineEditor
    term = getterminal()
    msg_enteridx = (
        term.bold_yellow(u'('),
        term.underline_yellow(u'0'),
        term.yellow(u'-'),
        term.underline_yellow(u'%d' % (len(locations) - 1,)),
        term.yellow(u','),
        term.underline_yellow('Escape'),
        term.bold_white(u':'),
        term.yellow('EXit'),
        term.bold_yellow(u')'), u' ',
        term.reverse_yellow(':'),)
    max_nwidth = len('%d' % (len(locations) - 1,))

    def disp_entry(num, loc):
        """ Display City, State.  """
        return u''.join((
            term.bold_yellow(u'['),
            u'%*d' % (max_nwidth, num),
            term.bold_yellow(u']'), u' ',
            term.yellow(loc['city']), u', ',
            term.yellow(loc['state']), u'\r\n',))
    echo(u'\r\n\r\n')
    lno = 3
    for num, loc in enumerate(locations):
        echo(disp_entry(num, loc))
        lno += 1
        if lno != 0 and (0 == lno % (term.height)):
            echo(term.yellow_reverse('--MORE--'))
            if getch() is None:
                break
            echo(u'\r\n')
            lno += 1
    idx = u''
    while True:
        echo(u'\r\n' + u''.join(msg_enteridx))
        idx = LineEditor(width=max_nwidth, content=idx).read()
        if idx is None or len(idx) == 0:
            return None
        try:
            int_idx = int(idx)
        except ValueError as err:
            echo(term.bold_red(u'\r\n%s' % (err,)))
            continue
        if int_idx < 0 or int_idx > len(locations) - 1:
            echo(term.bold_red(u'\r\nValue out of range'))
            continue
        return locations[int_idx]


def chose_location_lightbar(locations):
    """
    Lightbar pager for chosing a location.
    """
    from x84.bbs import getterminal, echo, Lightbar
    term = getterminal()
    fmt = u'%(city)s, %(state)s'
    lookup = dict([(loc['postal'], loc) for loc in locations])
    fullheight = min(term.height - 8, len(locations) + 2)
    fullwidth = min(75, int(term.width * .8))
    # shrink window to minimum width
    maxwidth = max([len(fmt % val) for val in lookup.values()]) + 2
    if maxwidth < fullwidth:
        fullwidth = maxwidth
    echo(u'\r\n' * fullheight)
    lightbar = Lightbar(height=fullheight,
                        width=fullwidth,
                        yloc=term.height - fullheight,
                        xloc=int((term.width / 2) - (fullwidth / 2)))
    lightbar.update([(key, fmt % val) for key, val in lookup.items()])
    lightbar.update([(key, fmt % val) for key, val in lookup.items()])
    lightbar.colors['border'] = term.yellow
    echo(lightbar.border())
    echo(lightbar.title(u''.join((
        term.yellow(u'-'), term.bold_white(u'[ '),
        term.bold_yellow('CitY'),
        term.bold_white(u', '),
        term.bold_yellow('StAtE'),
        term.bold_white(u' ]'), term.yellow(u'-'),))))
    echo(lightbar.footer(u''.join((
        term.yellow(u'-'), term.bold_black(u'( '),
        term.yellow_underline('Escape'), u':',
        term.yellow('EXit'),
        term.bold_black(u' )'), term.yellow(u'-'),))))
    lightbar.colors['highlight'] = term.yellow_reverse
    choice = lightbar.read()
    echo(lightbar.erase())
    return ((loc for loc in locations if choice == loc['postal']
             ).next() if choice is not None else choice)


def chose_location(locations):
    """
    Prompt user to chose a location.
    """
    from x84.bbs import getterminal, getsession, echo
    session, term = getsession(), getterminal()
    assert len(locations) > 0, (
        u'Cannot chose from empty list')
    msg_chosecity = (
        term.yellow(u'ChOSE A'),
        term.bold_yellow('CitY'),
        term.yellow_reverse(':'), u' ',)
    echo(u'\r\n\r\n')
    echo(u' '.join(msg_chosecity))
    if (session.user.get('expert', False) or 0 == term.number_of_colors):
        return chose_location_dummy(locations)
    return chose_location_lightbar(locations)


def location_prompt(location, msg='WEAthER'):
    """
    Prompt user to display weather or forecast.
    """
    from x84.bbs import getterminal, echo, getch
    term = getterminal()
    echo(u''.join((u'\r\n\r\n',
                   term.yellow(u'diSPlAY %s fOR ' % (msg,)),
                   term.bold('%(city)s, %(state)s' % location),
                   term.yellow(' ? '),
                   term.bold_yellow(u'['),
                   term.underline_yellow(u'yn'),
                   term.bold_yellow(u']'),
                   u': '),))
    while True:
        inp = getch()
        if inp is None or inp in (u'n', u'N', 'q', 'Q', term.KEY_EXIT):
            return False
        if inp in (u'y', u'Y', u' ', term.KEY_ENTER):
            return True


def disp_forecast(forecast):
    """ Display weather forecast.  """
    from x84.bbs import getterminal, echo, Ansi, getch
    term = getterminal()
    lno = 1
    echo(u'\r\n')
    for key in sorted(forecast.keys()):
        fcast = forecast[key]
        rstr = u''.join((
            term.bold_yellow_underline(fcast['DayCode']),
            u', ',
            u'/'.join(fcast['ObsDate'].split('/', 3)[0:2]),
            term.bold_underline(u':'), u' ',
            u'%s. ' % (
                term.bold(
                    fcast.get('TXT_Long', fcast.get('TXT_Short', u''))),),
            u'hiGH Of %sF, lOW Of %sF. ' % (
                term.bold(fcast.get('High_Temperature')),
                term.bold(fcast.get('Low_Temperature')),),))
        if 0.0 != float(fcast.get('Snow_Amount', '0.0')):
            rstr += u'SNOW AMOUNt %s iN., ' % (
                    term.yellow(fcast.get('Snow_Amount')),)
        if 0.0 != float(fcast.get('Rain_Amount', '0.0')):
            rstr += u'RAiN AMOUNt %s iN., ' % (
                    term.yellow(fcast.get('Snow_Amount')),)
        if 0.0 != float(fcast.get('Precip_Amount', '0.0')):
            rstr += u'PRECiPiTAtiON %s iN., ' % (
                    term.yellow(fcast.get('Precip_Amount')),)
        if 0 != int(fcast.get('TStorm_Prob', '0')):
            rstr += u'thUNdERStORM PRObAbilitY: %s, ' % (
                    term.yellow(fcast.get('TStorm_Prob')),)

        if 0 != len(fcast.get('WindDirection', u'')):
            rstr += u'WiNdS %s ' % (
                    term.yellow(fcast.get('WindDirection')),)
        if 0 != len(fcast.get('WindSpeed', u'')):
            rstr += u'At %sMPh, ' % (
                    term.yellow(fcast.get('WindSpeed')),)
        if 0 != len(fcast.get('WindGust', u'')):
            rstr += u'GUStS Of %sMPh. ' % (
                    term.yellow(fcast.get('WindGust')),)
            if 0 != len(fcast.get('Real_Feel_High', u'')):
                rstr += u'PROdUCiNG '
        if 0 != len(fcast.get('Real_Feel_High', u'')):
            rstr += u'A WiNdCHill HiGh Of %sF, lOW %sF. ' % (
                    term.yellow(
                        fcast.get('Real_Feel_High')),
                    term.yellow(
                        fcast.get('Real_Feel_Low', u'?')),)
        echo(u'\r\n')
        lno += 1
        lines = Ansi(rstr).wrap(min(60, int(term.width * .8))).splitlines()
        for line in lines:
            lno += 1
            echo(line + u'\r\n')
            if 0 == lno % (term.height - 1):
                echo(term.yellow_reverse('--MORE--'))
                if getch() is None:
                    return False
                echo(u'\r\n')


def disp_weather(weather):
    """ Display today's weather. """
    from x84.bbs import getterminal, echo, Ansi
    term = getterminal()
    rstr = u''.join((u'At ',
                     term.bold(u'%s%s' % (weather.get('City'),
                         u', %s' % (weather['State'],) if ('State' in weather
                             and 0 != len(weather['State'])) else u'',)),
                     term.underline(u':'), u' ',
                     u'%s. ' % (
                     term.bold_yellow(weather.get('WeatherText')),),
                     u'tEMPERAtURE Of %sF, ' % (
                     term.bold(weather.get('Temperature')),),
                     u'RElAtiVE hUMiditY %s. ' % (
                     term.yellow(weather.get('Humidity')),),
                     u'WiNdS %s At %s MPh' % (
                     term.yellow(weather.get('WindDirection')),
                     term.yellow(weather.get('WindSpeed'))),
                     u', PROdUCiNG A WiNdCHill Of %sF. ' % (
                     term.yellow(weather.get('RealFeel')),)
                     if (weather.get('RealFeel', weather.get('Temperature'))
                         != weather.get('Temperature'))
                     else u'. ',
                     u'ThE PRESSURE WAS %s iNChES ANd %s.' % (
                     term.yellow(weather.get('Pressure')),
                     term.yellow(weather.get('Pressure-state')),
                     ),))
    echo(u'\r\n\r\n')
    echo(Ansi(rstr).wrap(min(60, int(term.width * .8))))


def main():
    """ Main routine. """
    from x84.bbs import getsession, getterminal, echo, getch
    session, term = getsession(), getterminal()
    session.activity = 'Weather'

    echo(u'\r\n\r\n')
    location = session.user.get('location', dict())
    while True:
        search = location.get('postal', u'')
        disp_search_help()
        search = get_zipsearch(search)
        if search is None or 0 == len(search):
            return  # exit
        locations = do_search(search)
        if 0 != len(locations):
            location = (locations.pop() if 1 == len(locations)
                        else chose_location(locations) or dict())
        root = do_fetch(location.get('postal'))
        if root is None:
            return
        weather = parse_weather(root)
        #if False == location_prompt(location, 'WEAthER'):
        #    break
        disp_weather(weather)
        if False == location_prompt(location, 'fORECASt'):
            break
        forecast = parse_forecast(root)
        disp_forecast(forecast)
        echo(u'\r\n')
        echo(term.yellow_reverse('--ENd Of tRANSMiSSiON--'))
        getch()
        break

    if (sorted(location.items())
            != sorted(session.user.get('location', dict()).items())):
        echo(u''.join((u'\r\n\r\n',
                       term.yellow(u'SAVE lOCAtION'),
                       term.bold_yellow(' ('),
                       term.bold_black(u'PRiVAtE'),
                       term.bold_yellow(') '),
                       term.yellow('? '),
            term.bold_yellow(u'['),
            term.underline_yellow(u'yn'),
            term.bold_yellow(u']'),
            u': '),))
        while True:
            inp = getch()
            if inp is None or inp in (u'n', u'N', 'q', 'Q', term.KEY_EXIT):
                break
            if inp in (u'y', u'Y', u' ', term.KEY_ENTER):
                session.user['location'] = location
                break



"""
Database request handler for x/84 http://github.com/jquast/x84
"""
import threading
import os

FILELOCK = threading.Lock()


class DBHandler(threading.Thread):
    """
    This handler receives a "database command", in the form of a dictionary
    method name and its arguments, and the return value is sent to the session
    queue with the same 'event' name.
    """

    # pylint: disable=R0902
    #        Too many instance attributes (8/7)
    def __init__(self, queue, event, data):
        """ Arguments:
              inp_queue: parent input end of multiprocessing.Queue()
              event: database schema in form of string 'db-schema' or
                  'db=schema'. When '-' is used, the result is returned as a
                  single transfer. When '=', an iterable is yielded and the
                  data is transfered via the IPC Queue as a stream.
        """
        import x84.bbs.ini
        self.queue = queue
        self.event = event
        assert event[2] in ('-', '='), ('event name must match db[-=]event')
        self.iterable = event[2] == '='
        self.schema = event[3:]
        assert self.schema.isalnum() and os.path.sep not in self.schema, (
            'database schema "%s" must be alpha-numeric and not contain %s' % (
                self.schema, os.path.sep,))
        self.table = data[0]
        self.cmd = data[1]
        self.args = data[2]
        folder = x84.bbs.ini.CFG.get('system', 'datapath')
        self._tap_db = x84.bbs.ini.CFG.getboolean('session', 'tap_db')
        self.filepath = os.path.join(folder, '%s.sqlite3' % (self.schema,),)
        threading.Thread.__init__(self)

    def run(self):
        """
        Execute database command and return results to session queue.
        """
        import logging
        import sqlitedict
        logger = logging.getLogger(__name__)
        FILELOCK.acquire()
        # if the bbs is run as root, file ownerships become read-only
        # and db transactions will throw 'read-only database' errors,
        # exit earlier if we know that file permissions are to blame
        if not os.path.exists(os.path.dirname(self.filepath)):
            os.makedirs(os.path.dirname(self.filepath))
        assert os.access(os.path.dirname(self.filepath), os.F_OK|os.R_OK), (
                'Must have read+write+execute access to "%s" for database' % (
                    os.path.dirname(self.filepath),))
        if os.path.exists(self.filepath):
            assert os.access(self.filepath, os.F_OK|os.R_OK), (
                    'Must have read+write access to %s" for database' % (
                        self.filepath,))
        dictdb = sqlitedict.SqliteDict(
            filename=self.filepath, tablename=self.table, autocommit=True)
        FILELOCK.release()
        assert hasattr(dictdb, self.cmd), (
            "'%(cmd)s' not a valid method of <type 'dict'>" % self)
        func = getattr(dictdb, self.cmd)
        assert callable(func), (
            "'%(cmd)s' not a valid method of <type 'dict'>" % self)
        if self._tap_db:
            logger.debug('%s/%s%s', self.schema, self.cmd,
                    '(*%d)' % (len(self.args)) if len(self.args) else '()')

        # single value result,
        if not self.iterable:
            try:
                if 0 == len(self.args):
                    result = func()
                else:
                    result = func(*self.args)
            # pylint: disable=W0703
            #         Catching too general exception
            except Exception as err:
                # Pokemon exception; package & raise from session process,
                self.queue.send(('exception', err,))
                dictdb.close()
                logger.exception(err)
                return
            self.queue.send((self.event, result))
            dictdb.close()
            return

        # iterable value result,
        self.queue.send((self.event, (None, 'StartIteration'),))
        try:
            if 0 == len(self.args):
                for item in func():
                    self.queue.send((self.event, item,))
            else:
                for item in func(*self.args):
                    self.queue.send((self.event, item,))
        # pylint: disable=W0703
        #         Catching too general exception
        except Exception as err:
            # Pokemon exception; package & raise from session process,
            self.queue.send(('exception', err,))
            dictdb.close()
            logger.exception(err)
            return

        self.queue.send((self.event, (None, StopIteration,),))
        dictdb.close()
        return



"""
Translated from Markus Kuhn's C code at:

     http://www.cl.cam.ac.uk/~mgk25/ucs/wcwidth.c
"""
# This is an implementation of wcwidth() and wcswidth() (defined in
# IEEE Std 1002.1-2001) for Unicode.
#
# http://www.opengroup.org/onlinepubs/007904975/functions/wcwidth.html
# http://www.opengroup.org/onlinepubs/007904975/functions/wcswidth.html
#
# In fixed-width output devices, Latin characters all occupy a single
# "cell" position of equal width, whereas ideographic CJK characters
# occupy two such cells. Interoperability between terminal-line
# applications and (teletype-style) character terminals using the
# UTF-8 encoding requires agreement on which character should advance
# the cursor by how many cell positions. No established formal
# standards exist at present on which Unicode character shall occupy
# how many cell positions on character terminals. These routines are
# a first attempt of defining such behavior based on simple rules
# applied to data provided by the Unicode Consortium.
#
# For some graphical characters, the Unicode standard explicitly
# defines a character-cell width via the definition of the East Asian
# FullWidth (F), Wide (W), Half-width (H), and Narrow (Na) classes.
# In all these cases, there is no ambiguity about which width a
# terminal shall use. For characters in the East Asian Ambiguous (A)
# class, the width choice depends purely on a preference of backward
# compatibility with either historic CJK or Western practice.
# Choosing single-width for these characters is easy to justify as
# the appropriate long-term solution, as the CJK practice of
# displaying these characters as double-width comes from historic
# implementation simplicity (8-bit encoded characters were displayed
# single-width and 16-bit ones double-width, even for Greek,
# Cyrillic, etc.) and not any typographic considerations.
#
# Much less clear is the choice of width for the Not East Asian
# (Neutral) class. Existing practice does not dictate a width for any
# of these characters. It would nevertheless make sense
# typographically to allocate two character cells to characters such
# as for instance EM SPACE or VOLUME INTEGRAL, which cannot be
# represented adequately with a single-width glyph. The following
# routines at present merely assign a single-cell width to all
# neutral characters, in the interest of simplicity. This is not
# entirely satisfactory and should be reconsidered before
# establishing a formal standard in this area. At the moment, the
# decision which Not East Asian (Neutral) characters should be
# represented by double-width glyphs cannot yet be answered by
# applying a simple rule from the Unicode database content. Setting
# up a proper standard for the behavior of UTF-8 character terminals
# will require a careful analysis not only of each Unicode character,
# but also of each presentation form, something the author of these
# routines has avoided to do so far.
#
# http://www.unicode.org/unicode/reports/tr11/
#
# Markus Kuhn -- 2003-05-20 (Unicode 4.0)
#
# Permission to use, copy, modify, and distribute this software
# for any purpose and without fee is hereby granted. The author
# disclaims all warranties with regard to this software.
#
# Latest version: http://www.cl.cam.ac.uk/~mgk25/ucs/wcwidth.c


def _bisearch(ucs, table):
    " auxiliary function for binary search in interval table "
    lbound = 0
    ubound = len(table) - 1

    if ucs < table[0][0] or ucs > table[ubound][1]:
        return 0
    while ubound >= lbound:
        mid = (lbound + ubound) / 2
        if ucs > table[mid][1]:
            lbound = mid + 1
        elif ucs < table[mid][0]:
            ubound = mid - 1
        else:
            return 1

    return 0

# The following two functions define the column width of an ISO 10646
# character as follows:
#
#    - The null character (U+0000) has a column width of 0.
#
#    - Other C0/C1 control characters and DEL will lead to a return
#      value of -1.
#
#    - Non-spacing and enclosing combining characters (general
#      category code Mn or Me in the Unicode database) have a
#      column width of 0.
#
#    - SOFT HYPHEN (U+00AD) has a column width of 1.
#
#    - Other format characters (general category code Cf in the Unicode
#      database) and ZERO WIDTH SPACE (U+200B) have a column width of 0.
#
#    - Hangul Jamo medial vowels and final consonants (U+1160-U+11FF)
#      have a column width of 0.
#
#    - Spacing characters in the East Asian Wide (W) or East Asian
#      Full-width (F) category as defined in Unicode Technical
#      Report #11 have a column width of 2.
#
#    - All remaining characters (including all printable
#      ISO 8859-1 and WGL4 characters, Unicode control characters,
#      etc.) have a column width of 1.
#
# This implementation assumes that wchar_t characters are encoded
# in ISO 10646.

# sorted list of non-overlapping intervals of non-spacing characters
# generated by "uniset +cat=Me +cat=Mn +cat=Cf -00AD +1160-11FF +200B c"

_COMBINING = [
    (0x0300, 0x0357), (0x035D, 0x036F), (0x0483, 0x0486),
    (0x0488, 0x0489), (0x0591, 0x05A1), (0x05A3, 0x05B9),
    (0x05BB, 0x05BD), (0x05BF, 0x05BF), (0x05C1, 0x05C2),
    (0x05C4, 0x05C4), (0x0600, 0x0603), (0x0610, 0x0615),
    (0x064B, 0x0658), (0x0670, 0x0670), (0x06D6, 0x06E4),
    (0x06E7, 0x06E8), (0x06EA, 0x06ED), (0x070F, 0x070F),
    (0x0711, 0x0711), (0x0730, 0x074A), (0x07A6, 0x07B0),
    (0x0901, 0x0902), (0x093C, 0x093C), (0x0941, 0x0948),
    (0x094D, 0x094D), (0x0951, 0x0954), (0x0962, 0x0963),
    (0x0981, 0x0981), (0x09BC, 0x09BC), (0x09C1, 0x09C4),
    (0x09CD, 0x09CD), (0x09E2, 0x09E3), (0x0A01, 0x0A02),
    (0x0A3C, 0x0A3C), (0x0A41, 0x0A42), (0x0A47, 0x0A48),
    (0x0A4B, 0x0A4D), (0x0A70, 0x0A71), (0x0A81, 0x0A82),
    (0x0ABC, 0x0ABC), (0x0AC1, 0x0AC5), (0x0AC7, 0x0AC8),
    (0x0ACD, 0x0ACD), (0x0AE2, 0x0AE3), (0x0B01, 0x0B01),
    (0x0B3C, 0x0B3C), (0x0B3F, 0x0B3F), (0x0B41, 0x0B43),
    (0x0B4D, 0x0B4D), (0x0B56, 0x0B56), (0x0B82, 0x0B82),
    (0x0BC0, 0x0BC0), (0x0BCD, 0x0BCD), (0x0C3E, 0x0C40),
    (0x0C46, 0x0C48), (0x0C4A, 0x0C4D), (0x0C55, 0x0C56),
    (0x0CBC, 0x0CBC), (0x0CBF, 0x0CBF), (0x0CC6, 0x0CC6),
    (0x0CCC, 0x0CCD), (0x0D41, 0x0D43), (0x0D4D, 0x0D4D),
    (0x0DCA, 0x0DCA), (0x0DD2, 0x0DD4), (0x0DD6, 0x0DD6),
    (0x0E31, 0x0E31), (0x0E34, 0x0E3A), (0x0E47, 0x0E4E),
    (0x0EB1, 0x0EB1), (0x0EB4, 0x0EB9), (0x0EBB, 0x0EBC),
    (0x0EC8, 0x0ECD), (0x0F18, 0x0F19), (0x0F35, 0x0F35),
    (0x0F37, 0x0F37), (0x0F39, 0x0F39), (0x0F71, 0x0F7E),
    (0x0F80, 0x0F84), (0x0F86, 0x0F87), (0x0F90, 0x0F97),
    (0x0F99, 0x0FBC), (0x0FC6, 0x0FC6), (0x102D, 0x1030),
    (0x1032, 0x1032), (0x1036, 0x1037), (0x1039, 0x1039),
    (0x1058, 0x1059), (0x1160, 0x11FF), (0x1712, 0x1714),
    (0x1732, 0x1734), (0x1752, 0x1753), (0x1772, 0x1773),
    (0x17B4, 0x17B5), (0x17B7, 0x17BD), (0x17C6, 0x17C6),
    (0x17C9, 0x17D3), (0x17DD, 0x17DD), (0x180B, 0x180D),
    (0x18A9, 0x18A9), (0x1920, 0x1922), (0x1927, 0x1928),
    (0x1932, 0x1932), (0x1939, 0x193B), (0x200B, 0x200F),
    (0x202A, 0x202E), (0x2060, 0x2063), (0x206A, 0x206F),
    (0x20D0, 0x20EA), (0x302A, 0x302F), (0x3099, 0x309A),
    (0xFB1E, 0xFB1E), (0xFE00, 0xFE0F), (0xFE20, 0xFE23),
    (0xFEFF, 0xFEFF), (0xFFF9, 0xFFFB), (0x1D167, 0x1D169),
    (0x1D173, 0x1D182), (0x1D185, 0x1D18B), (0x1D1AA, 0x1D1AD),
    (0xE0001, 0xE0001), (0xE0020, 0xE007F), (0xE0100, 0xE01EF)
]


def wcwidth(char):
    """
    Return the width in character cells of the Unicode character
    whose code is c
    """

    ucs = ord(char)
    # test for 8-bit control characters
    if ucs == 0:
        return 0
    if ucs < 32 or (ucs >= 0x7f and ucs < 0xa0):
        return -1

    # binary search in table of non-spacing characters
    if _bisearch(ucs, _COMBINING):
        return 0

    # if we arrive here, ucs is not a combining or C0/C1 control character

    return 1 + \
        (ucs >= 0x1100 and
        (ucs <= 0x115f or                  # Hangul Jamo init. consonants
         ucs == 0x2329 or ucs == 0x232a or
        (ucs >= 0x2e80 and ucs <= 0xa4cf and
         ucs != 0x303f) or                # CJK ... Yi
        (ucs >= 0xac00 and ucs <= 0xd7a3) or  # Hangul Syllables
        (ucs >= 0xf900 and ucs <= 0xfaff) or  # CJK Compatibility Ideographs *
        (ucs >= 0xfe30 and ucs <= 0xfe6f) or  # CJK Compatibility Forms
        (ucs >= 0xff00 and ucs <= 0xff60) or  # Fullwidth Forms
        (ucs >= 0xffe0 and ucs <= 0xffe6) or
        (ucs >= 0x20000 and ucs <= 0x2fffd) or
        (ucs >= 0x30000 and ucs <= 0x3fffd)))


def wcswidth(pwcs):

    """
    Return the width in character cells of the unicode string pwcs,
    or -1 if the string contains non-printable characters.
    """

    width = 0
    for char in pwcs:
        wcw = wcwidth(char)
        if wcw < 0:
            return -1
        else:
            width += wcw
    return width


# sorted list of non-overlapping intervals of East Asian Ambiguous
# characters, generated by "uniset +WIDTH-A -cat=Me -cat=Mn -cat=Cf c"
_AMBIGUOUS = [
    (0x00A1, 0x00A1), (0x00A4, 0x00A4), (0x00A7, 0x00A8),
    (0x00AA, 0x00AA), (0x00AE, 0x00AE), (0x00B0, 0x00B4),
    (0x00B6, 0x00BA), (0x00BC, 0x00BF), (0x00C6, 0x00C6),
    (0x00D0, 0x00D0), (0x00D7, 0x00D8), (0x00DE, 0x00E1),
    (0x00E6, 0x00E6), (0x00E8, 0x00EA), (0x00EC, 0x00ED),
    (0x00F0, 0x00F0), (0x00F2, 0x00F3), (0x00F7, 0x00FA),
    (0x00FC, 0x00FC), (0x00FE, 0x00FE), (0x0101, 0x0101),
    (0x0111, 0x0111), (0x0113, 0x0113), (0x011B, 0x011B),
    (0x0126, 0x0127), (0x012B, 0x012B), (0x0131, 0x0133),
    (0x0138, 0x0138), (0x013F, 0x0142), (0x0144, 0x0144),
    (0x0148, 0x014B), (0x014D, 0x014D), (0x0152, 0x0153),
    (0x0166, 0x0167), (0x016B, 0x016B), (0x01CE, 0x01CE),
    (0x01D0, 0x01D0), (0x01D2, 0x01D2), (0x01D4, 0x01D4),
    (0x01D6, 0x01D6), (0x01D8, 0x01D8), (0x01DA, 0x01DA),
    (0x01DC, 0x01DC), (0x0251, 0x0251), (0x0261, 0x0261),
    (0x02C4, 0x02C4), (0x02C7, 0x02C7), (0x02C9, 0x02CB),
    (0x02CD, 0x02CD), (0x02D0, 0x02D0), (0x02D8, 0x02DB),
    (0x02DD, 0x02DD), (0x02DF, 0x02DF), (0x0391, 0x03A1),
    (0x03A3, 0x03A9), (0x03B1, 0x03C1), (0x03C3, 0x03C9),
    (0x0401, 0x0401), (0x0410, 0x044F), (0x0451, 0x0451),
    (0x2010, 0x2010), (0x2013, 0x2016), (0x2018, 0x2019),
    (0x201C, 0x201D), (0x2020, 0x2022), (0x2024, 0x2027),
    (0x2030, 0x2030), (0x2032, 0x2033), (0x2035, 0x2035),
    (0x203B, 0x203B), (0x203E, 0x203E), (0x2074, 0x2074),
    (0x207F, 0x207F), (0x2081, 0x2084), (0x20AC, 0x20AC),
    (0x2103, 0x2103), (0x2105, 0x2105), (0x2109, 0x2109),
    (0x2113, 0x2113), (0x2116, 0x2116), (0x2121, 0x2122),
    (0x2126, 0x2126), (0x212B, 0x212B), (0x2153, 0x2154),
    (0x215B, 0x215E), (0x2160, 0x216B), (0x2170, 0x2179),
    (0x2190, 0x2199), (0x21B8, 0x21B9), (0x21D2, 0x21D2),
    (0x21D4, 0x21D4), (0x21E7, 0x21E7), (0x2200, 0x2200),
    (0x2202, 0x2203), (0x2207, 0x2208), (0x220B, 0x220B),
    (0x220F, 0x220F), (0x2211, 0x2211), (0x2215, 0x2215),
    (0x221A, 0x221A), (0x221D, 0x2220), (0x2223, 0x2223),
    (0x2225, 0x2225), (0x2227, 0x222C), (0x222E, 0x222E),
    (0x2234, 0x2237), (0x223C, 0x223D), (0x2248, 0x2248),
    (0x224C, 0x224C), (0x2252, 0x2252), (0x2260, 0x2261),
    (0x2264, 0x2267), (0x226A, 0x226B), (0x226E, 0x226F),
    (0x2282, 0x2283), (0x2286, 0x2287), (0x2295, 0x2295),
    (0x2299, 0x2299), (0x22A5, 0x22A5), (0x22BF, 0x22BF),
    (0x2312, 0x2312), (0x2460, 0x24E9), (0x24EB, 0x254B),
    (0x2550, 0x2573), (0x2580, 0x258F), (0x2592, 0x2595),
    (0x25A0, 0x25A1), (0x25A3, 0x25A9), (0x25B2, 0x25B3),
    (0x25B6, 0x25B7), (0x25BC, 0x25BD), (0x25C0, 0x25C1),
    (0x25C6, 0x25C8), (0x25CB, 0x25CB), (0x25CE, 0x25D1),
    (0x25E2, 0x25E5), (0x25EF, 0x25EF), (0x2605, 0x2606),
    (0x2609, 0x2609), (0x260E, 0x260F), (0x2614, 0x2615),
    (0x261C, 0x261C), (0x261E, 0x261E), (0x2640, 0x2640),
    (0x2642, 0x2642), (0x2660, 0x2661), (0x2663, 0x2665),
    (0x2667, 0x266A), (0x266C, 0x266D), (0x266F, 0x266F),
    (0x273D, 0x273D), (0x2776, 0x277F), (0xE000, 0xF8FF),
    (0xFFFD, 0xFFFD), (0xF0000, 0xFFFFD), (0x100000, 0x10FFFD)
]

# The following functions are the same as mk_wcwidth() and
# mk_wcwidth_cjk(), except that spacing characters in the East Asian
# Ambiguous (A) category as defined in Unicode Technical Report #11
# have a column width of 2. This variant might be useful for users of
# CJK legacy encodings who want to migrate to UCS without changing
# the traditional terminal character-width behaviour. It is not
# otherwise recommended for general use.


def wcwidth_cjk(ucs):
    """ As wcwidth above, but spacing characters in the East Asian
    Ambiguous (A) category as defined in Unicode Technical Report #11
    have a column width of 2.
    """
    if _bisearch(ucs, _AMBIGUOUS):
        return 2
    else:
        return wcwidth(ucs)


def wcswidth_cjk(pwcs):
    """ As wcswidth above, but spacing characters in the East Asian
    Ambiguous (A) category as defined in Unicode Technical Report #11
    have a column width of 2.
    """
    width = 0
    for char in pwcs:
        wcw = wcwidth_cjk(char)
        if wcw < 0:
            return -1
        else:
            width += wcw
    return width


#####################################################################
def _measure_string(ucs, length):
    " helper function for padding "
    tot = 0
    idx = 0
    while tot < length and idx < len(ucs):
        tot += wcswidth(ucs[idx])
        idx += 1
    return (ucs[:idx], tot)


def rpadstring(ucs, length, padchar=' '):
    """ Right-pad a Unicode string with padchar so that its width in
    character cells is length.  Padchar must be of width 1.  The string
    is truncated if it is too long."""
    pstr, plen = _measure_string(ucs, length)
    if plen > length:
        return pstr[:-1] + padchar
    elif plen < length:
        return pstr + (padchar * (length - plen))
    else:
        return pstr


def truncatestring(ucs, length):
    """ Truncate a Unicode string so that its length is as long as it
    can be without exceeding length."""

    pstr, plen = _measure_string(ucs, length)

    if plen > length:
        return pstr[:-1]
    else:
        return pstr
